
	/*************/
	/* Variables */
	/*************/

var	personality;

var	builderGroup, oilerGroup, battleGroup = [], vtolGroup = [], aaVtolGroup, abVtolGroup, defendGroup, harassGroup, aaGroup;
var	groupInfo = [];
var	enemyInfo = [];
var	derrickStats = [];
var	hotspots = []; //These are position objects that record where attacks on AI-controlled objects take place.

var warnVtols=false;

const	STATUS_NORMAL=0, STATUS_PANIC=1, STATUS_EARLYGAME=2, STATUS_UNSTUCK=3, STATUS_HOMELESS=4;
const	TARGET_DROID=0, TARGET_STRUCTURE=1;
var	status=STATUS_NORMAL;
var	statusX, statusY, statusTime;
var	lastArtyAttackX=-1, lastArtyAttackY=-1;
var	lastAirAttackX=-1, lastAirAttackY=-1;
var	attackedByChoppers; //Used in case the AI's base is under attack by Scavenger Choppers (from the Ultimate Scavenger Mod)
var	basicBuildingQueue; //Used to help keep the AI from traveling a long way to build base structures.
var	basicBuildingLoc;

var	global=this;
var	basePosition;

var forcedResearchPath = [];

var beaconInfo = [];

if (typeof(scavengerPlayer)=="undefined") // HACK: until scavengerPlayer is exposed to scripts.
	scavengerPlayer=Math.max(7,maxPlayers)

const	COMP_PROPULSION=3; // HACK: define it here until these constants are exposed to scripts
const	COMP_BODY=1; 
const	COMP_WEAPON=8;

const DROID_CYBORG_CONSTRUCT=10; // HACK: see also http://forums.wz2100.net/viewtopic.php?f=35&t=10341&p=114181#p114184

const DORDER_CIRCLE=40;

	/**************/
	/* Procedures */
	/**************/

// creates a new instance of group info
function constructGroupInfo(enemy) {
	this.enemy=enemy;
	this.targetX=-1;
	this.targetY=-1;
	this.targetType=TARGET_STRUCTURE;
	this.targetAge=0;
	this.targetStatType=0;
	this.delayregroup=0;
	this.delaystaticregroup=0;
	this.lastX=-1;
	this.lastY=-1;
	this.idleTime=0;
	this.helpX=-1;
	this.helpY=-1;
	this.helpPlayer=-1;
	this.helpLastUpdate=gameTime;
}

// creates a new instance of enemy info
function constructEnemyInfo() {
	this.tank = 0;
	this.borg = 0;
	this.vtol = 0;
	this.defs = 0;
	this.aa = 0;
	this.lastUpdate = 0;
	this.tanksAtBase = 0;
	this.reachable = true;
}

function constructBeaconInfo() {
	this.x = -1;
	this.y = -1;
	this.exists = 0;
	this.lastUpdate = 0;
}

//Creates a new instance of a hotspot position object.
function constructHotspotInfo(x,y) {
	this.x = x;
	this.y = y;
	this.type = POSITION;
	this.lifespan = 10;
}

//Checks if a position is located near an active hotspot object.
function isNearHotspot(x,y) {
	for (var i=0; i<hotspots.length; ++i)
		if (hotspots[i].lifespan>0)
			if (distBetweenTwoPoints(x, y, hotspots[i].x, hotspots[i].y) <= HOTSPOT_SIZE)
				return true;
	return false;
}

//Checks if a position is located in the main base.
function isNearBase(x,y) {
	return (distBetweenTwoPoints(x, y, basePosition.x, basePosition.y) <= BASE_SIZE);
}

// finds the nearest object within a given array of objects.
function findNearestObject(x,y,objectlist) {
	var minDist=Infinity;
	var nearestObj;
	for (var i=0; i<objectlist.length; ++i) {
		var dist=distBetweenTwoPoints(x,y,objectlist[i].x,objectlist[i].y);
		if (dist<minDist) {
			minDist=dist;
			nearestObj=objectlist[i];
		}
	}
	return nearestObj;
}

//Sorts an array of objects based on distance to the given coordinates.
function sortByDistance(x,y,objectlist) {
	var sortedList=[]
	var minDist=Infinity;
	for (var i=0; i<objectlist.length; ++i) {
		minDist=Infinity;
		for (var j=0; j<objectlist.length; ++j) {
			if (j<i)
				continue;
			var dist=distBetweenTwoPoints(x,y,objectlist[j].x,objectlist[j].y);
			if (dist<minDist) {
				minDist=dist;
				sortedList[i]=objectlist[j];
			}
		}
	}
	return sortedList;
}

// this AI doesn't cheat yet
function cheat() {
	switch(difficulty) {
		case INSANE:
			break;
		case HARD:
			break;
		case MEDIUM:
			break;
		case EASY:
			break;
	}
}

// makes attack groups bigger
function becomeHarder() {
	if (BECOMES_HARDER) {
		var enemies=countEnemies()+1;
		MAX_WARRIORS = Math.floor(Math.min(150/enemies,13+gameTime/180000));
		NUM_WARRIORS = Math.floor(Math.min(150/enemies,9+gameTime/180000));
		MIN_WARRIORS = Math.floor(Math.min(NUM_WARRIORS*0.8,5+gameTime/180000));
		MAX_VTOLS = Math.floor(Math.min(Math.min(40,9+2*HIGH_TECH_START+gameTime/360000),150-150/enemies));
		NUM_VTOLS = Math.floor(Math.min(Math.min(36,5+2*HIGH_TECH_START+gameTime/360000),MAX_VTOLS));
		MIN_VTOLS = Math.floor(Math.min(Math.min(28,1+2*HIGH_TECH_START+gameTime/360000),NUM_VTOLS*0.8));
	}
}

// call for help, recognizing how to do that for different sorts of allies
function callForHelp() {
	if (typeof(addBeacon)!="undefined") {
		addBeacon(basePosition.x, basePosition.y, ALLIES)
	}
	if (typeof(chat)=="undefined")
		return;
	chat(ALLIES,"Could use some help here!");
	for (var i=0; i<maxPlayers; ++i) if (allianceExistsBetween(me,i) && playerData[i].isAI && i!=me) {
		// Nexus or Semperfi AI
		if (playerData[i].name.indexOf("Nexus")>-1 || playerData[i].name.indexOf("Semperfi")>-1) {
			chat(i,"help");
			continue;
		}
		// Another instance of NullBot
		if (playerData[i].name.indexOf("NullBot")>-1) {
			chat(i,"nb help");
			continue;
		}
	}
}

// order the strongest free group to help another player
function sendHelp(player) {
	var maxCount=0,maxIdx=0;
	// send the biggest combat group
	for (var i=0; i<NUM_GROUPS; ++i) if (groupInfo[i].helpPlayer == -1) {
		var n=enumGroup(battleGroup[i]).length;
		if (n > maxCount) {
			maxCount = n;
			maxIdx = i;
		}
	}
	if (maxCount < MIN_WARRIORS) {
		return "Sorry, don't have any free forces to send for help!";
	}
	groupInfo[maxIdx].helpX=beaconInfo[player].x;
	groupInfo[maxIdx].helpY=beaconInfo[player].y;
	groupInfo[maxIdx].helpPlayer=player;
	groupInfo[maxIdx].helpLastUpdate=gameTime;
	return "Sending "+maxCount+" droids of combat group "+maxIdx+" to help you at ("+groupInfo[maxIdx].helpX+","+groupInfo[maxIdx].helpY+")."	
}

// no longer send help to a certain player
function dontSendHelp(player) {
	for (var i=0; i<NUM_GROUPS; ++i) if (groupInfo[i].helpPlayer == player)
		battleGroup[i].helpPlayer=-1;
	return "You're welcome!"
}

// give a truck to a certain player
function giveTruck(player) {
	var list=enumDroid(me,DROID_CONSTRUCT);
	if (list.length<=0) {
		return "Sorry, I have no trucks.";
	}
	var droid;
	if (beaconInfo[player].exists) { // if the player has posted a beacon, give a truck nearest to the beacon
		var minDist=Infinity;
		for (var i=0; i<list.length; ++i) {
			var dist=distBetweenTwoPoints(beaconInfo[player].x,beaconInfo[player].y,list[i].x,list[i].y);
			if (dist<minDist) {
				minDist=dist;
				droid=list[i];
			}
		}
	} else // otherwise give a random truck
		droid=list[0];
	addBeacon(droid.x,droid.y,player);
	if (donateObject(droid,player))
		return "You can use this one.";
	else
		return "Sorry, you are at limits already.";
}

// give power to a certain player
// (shouldn't normally work, cause the AI most likely has 0 power)
function giveMoney(player) {
	var power=Math.round(myPower()/3);
	donatePower(power,player);
	return "Donated "+power+" power!";
}

// makes sure we have at least one operational structure of that type
function iHaveStruct(stat) {
	var list=enumStruct(me,stat)
	for (i=0; i<list.length; ++i)
		if (list[i].status==BUILT)
			return true;
	return false;
}

// checks if we can produce hovertanks
function iHaveHovers() {
	return componentAvailable(COMP_PROPULSION,"hover01");	
}

// checks if it's time to build sensors all over the place
function iHaveArty() {
	var art = personality.apArtillery.concat(personality.atArtillery,personality.abArtillery)
	var success = false;
	for (var i=0; i<art.length; ++i) {
		success = isStructureAvailable(art[i],me);
		if (success)
			break;
	}
	return success;
}

// checks if at least one generator is up and running when necessary
function dontRunOutOfPower() {
	// check for T2>= start and don't hang with no oil because of this procedure until a generator is built
	if (HIGH_TECH_START==1) {
		return iHaveStruct(generator) && iHaveStruct(derrick);
	}
	return true;
}

// checks if a structure is a dedicated anti-air defense
function isDedicatedAA(struct) {
	if (typeof(struct.canHitAir)!="undefined" && typeof(struct.canHitGround)!="undefined") {
		return struct.canHitAir && !struct.canHitGround;
	} else {
		// some workaround since we lack the functions used above in v3.1
		if (struct.name.indexOf("AA")>-1) 
			return true;
		if (struct.name.indexOf("SAM")>-1) 
			return true;
		if (struct.name.indexOf("Tornado")>-1) 
			return true;
		if (struct.name.indexOf("Hurricane")>-1) 
			return true;
		if (struct.name.indexOf("Cyclone")>-1) 
			return true;
		if (struct.name.indexOf("Whirlwhind")>-1) 
			return true;
		if (struct.name.indexOf("Dust Devil")>-1) 
			return true;
		if (struct.name.indexOf("Sandstorm")>-1) 
			return true;
		if (struct.name.indexOf("Sunburst")>-1) 
			return true;
		if (struct.name.indexOf("Cherub")>-1) 
			return true;
		if (struct.name.indexOf("Solar Flare")>-1) 
			return true;
		if (struct.name.indexOf("Sunstorm")>-1) 
			return true;
		if (struct.name.indexOf("Avenger")>-1) 
			return true;
		if (struct.name.indexOf("Vindicator")>-1) 
			return true;
		if (struct.name.indexOf("Stormbringer")>-1) 
			return true;
		if (struct.name.indexOf("Tempest")>-1) 
			return true;
	}
	return false;
}

function isLongRangeArtillery(struct) {
	if (typeof(struct.range)!="undefined") {
		return (struct.range > ARTILLERY_RANGE);
	} else {
		// some workaround since we lack the function used above in v3.1
		if (struct.name.indexOf("Howitzer")>-1) 
			return true;
		if (struct.name.indexOf("TBM")>-1) 
			return true;
		if (struct.name.indexOf("Ground Shaker")>-1) 
			return true;
		if (struct.name.indexOf("Hellstorm")>-1) 
			return true;
		if (struct.name.indexOf("Ragnarok")>-1) 
			return true;
		if (struct.name.indexOf("Meteor Launcher")>-1) 
			return true;
		if (struct.name.indexOf("Ripple")>-1) 
			return true;
		if (struct.name.indexOf("Archangel")>-1) 
			return true;
		if (struct.name.indexOf("Shellstorm")>-1) 
			return true;
		if (struct.name.indexOf("Sunrider")>-1) 
			return true;
		if (struct.name.indexOf("Apollo")>-1) 
			return true;
		if (struct.name.indexOf("Ballista")>-1) 
			return true;
	}
	return false;
}

// count tank/borg balance of a particular enemy
// enemy=-1 will count all enemies as one
function enemyCountBalance(enemy) {
	if (difficulty == EASY || difficulty == MEDIUM)
		return {
			tank:12,
			borg:12,
			vtol:3,
			defs:3,
			aa:5
		};
	// the rest involves deity cheating
	if (typeof(enemyInfo[enemy])!="undefined") {
		if (gameTime - enemyInfo[enemy].lastUpdate < 60000)
			return enemyInfo[enemy];
	} else 
		enemyInfo[enemy]=new Object();
	enemyInfo[enemy].tank=0;
	enemyInfo[enemy].borg=0;
	enemyInfo[enemy].vtol=0;
	enemyInfo[enemy].defs=0;
	enemyInfo[enemy].aa=0;
	var list=cachedEnumEnemies(enemy);
	for (var i=0; i<list.length; ++i) {
		var obj=list[i];
		if (obj.player == enemy || (enemy == -1 && isAnEnemy(obj.player))) {
			if (obj.type == DROID) {
				if (obj.isVTOL) {
					++enemyInfo[enemy].vtol;
				} else if (obj.droidType == DROID_WEAPON) {
					++enemyInfo[enemy].tank;
				} else if (obj.droidType == DROID_CYBORG) {
					++enemyInfo[enemy].borg;
				}
			} else if (obj.type == STRUCTURE) {
				if (obj.stattype == DEFENSE) {
					if (isDedicatedAA(obj)) {
						++enemyInfo[enemy].aa;
					} else {
						++enemyInfo[enemy].defs;
					}
				}
			}
		}
	}
	enemyInfo[enemy].lastUpdate = gameTime;
	return enemyInfo[enemy];
}

// counts AP/AT balance of a battle group
function groupCountBalance(gr) {
	var list=enumGroup(battleGroup[gr]);
	var ap=0, at=0, ab=0, aa=0, se=0, cb=0, de=0, emp=0;
	var size=1;
	for (var i=0; i<list.length; ++i) {
		//Different size units often perform their roles with different degrees of effectiveness.
		//Note that this doesn't necessarily hold true for sensor, cb, and detector tanks.
		if (list[i].name.indexOf("Cyborg") > -1)
			size=1;
		else if (list[i].name.indexOf("Light") > -1)
			size=2;
		else if (list[i].name.indexOf("Medium") > -1)
			size=4;
		else if (list[i].name.indexOf("Heavy") > -1)
			size=8;
		else if (list[i].name.indexOf("Super") > -1)
			size=16;
		//Now that we have the unit's size, let's find out which role the unit belongs to.
		if (list[i].name.indexOf("AP") > -1)
			ap+=size;
		else if (list[i].name.indexOf("AT") > -1)
			at+=size;
		else if (list[i].name.indexOf("AB") > -1)
			ab+=size;
		else if (list[i].name.indexOf("AA") > -1)
			aa+=size;
		else if (list[i].name.indexOf("Sensor") > -1)
			se+=1;
		else if (list[i].name.indexOf("CB") > -1)
			cb+=1;
		else if (list[i].name.indexOf("Detector") > -1)
			de+=1;
		else if (list[i].name.indexOf("EMP") > -1)
			emp+=1;
	}
	return {ap:ap,at:at,ab:ab,aa:aa,se:se,cb:cb,de:de,emp:emp};
}

function groupCountBalanceVtol(gr) {
	var list=enumGroup(vtolGroup[gr]);
	var ap=0, at=0, ab=0, aa=0, se=0, cb=0, de=0, emp=0;
	var size=1;
	for (var i=0; i<list.length; ++i) {
		//Different size units often perform their roles with different degrees of effectiveness.
		//Note that this doesn't necessarily hold true for sensor, cb, and detector tanks.
		if (list[i].name.indexOf("Light") > -1)
			size=2;
		else if (list[i].name.indexOf("Medium") > -1)
			size=4;
		else if (list[i].name.indexOf("Heavy") > -1)
			size=8;
		else if (list[i].name.indexOf("Super") > -1)
			size=16;
		//Now that we have the unit's size, let's find out which role the unit belongs to.
		if (list[i].name.indexOf("AP") > -1)
			ap+=size;
		else if (list[i].name.indexOf("AT") > -1)
			at+=size;
		else if (list[i].name.indexOf("AB") > -1)
			ab+=size;
		else if (list[i].name.indexOf("AA") > -1)
			aa+=size;
		else if (list[i].name.indexOf("Sensor") > -1)
			se+=1;
		else if (list[i].name.indexOf("CB") > -1)
			cb+=1;
		else if (list[i].name.indexOf("Detector") > -1)
			de+=1;
		else if (list[i].name.indexOf("EMP") > -1)
			emp+=1;
	}
	return {ap:ap,at:at,ab:ab,aa:aa,se:se,cb:cb,de:de,emp:emp};
}

// the heart of the AI adaptation algorithm
function adaptationMatrix(ap, at, aa, ab, tank, borg, vtol, defs) {
	if (difficulty == EASY || difficulty == MEDIUM)
		return {ap:4,at:4,aa:1,ab:1};
	// the rest involves deity cheating
	var newAP, newAT, newAA, newAB;
	var totalMe=ap+at+aa+ab; 
	var totalEnemy=tank+borg+vtol+defs;
	if (totalMe==0 || totalEnemy==0)
		return {ap:1,at:1,aa:0,ab:0};
	newAP = 100*(borg/(totalEnemy*4) - ap/totalMe); //Heavy Cyborgs in Contingency have roughly 1/4th of the health, price, and build points of a light tank.
	newAT = 100*(tank/totalEnemy - at/totalMe);
	newAA = 100*(vtol/totalEnemy - 2*aa/totalMe)+3+4*(warnVtols==true);
	newAB = 100*(defs/totalEnemy - 2*ab/totalMe);
	newAP = Math.floor(newAP); if (newAP < 2) newAP=2;
	newAT = Math.floor(newAT); if (newAT < 1) newAT=1;
	newAA = Math.floor(newAA); if (newAA < 0) newAA=0;
	newAB = Math.floor(newAB); if (newAB < 0) newAB=0;
	return {ap:newAP,at:newAT,aa:newAA,ab:newAB};
}

// adapt to enemy choices and game situation
function adapt() {
	var seaEnemies = 0, landEnemies = 0;
	for (var i=0; i<maxPlayers; ++i) if (isAnEnemy(i)) {
		if (enemyInfo[i].reachable)
			++landEnemies;
		else
			++seaEnemies;
	}
	if (RATE_TANK > 0) {
		RATE_TANK = landEnemies;
		RATE_HOVER = seaEnemies;
	} else {
		RATE_TANK = 0; // don't use tanks once we switched to pure hovers at least once (kind of the safe way)
		RATE_HOVER = 1;
	}
	if (difficulty == EASY || difficulty == MEDIUM) {
		RATE_AP = 12;
		RATE_AT = 12;
		RATE_AA = 3;
		RATE_AB = 3;
	} else { // the rest involves deity cheating
		var enemyBorgCount = 0, enemyTankCount = 0, enemyVtolCount = 0, enemyBuildingCount = 0;
		var ret=enemyInfo[-1];
		if (typeof(ret)=="undefined") { // usually happens on game load
			callUpdateEnemyInfo();
			ret=enemyInfo[-1];
		}
		enemyBorgCount+=ret.borg;
		enemyTankCount+=ret.tank;
		enemyVtolCount+=ret.vtol;
		enemyBuildingCount+=ret.defs;
		for (var i=0; i<maxPlayers; ++i) if (isAnEnemy(i)) {
			enemyVtolCount+=2*enumStruct(i,vtolfac).length;
		}
		//Count Scavenger Choppers as well.
		enemyVtolCount+=enumDroid(scavengerPlayer,DROID_TRANSPORTER).length;
		
		var myATCount = 0, myAPCount = 0, myAACount = 0, myABCount = 0;
		for (var i=0; i<NUM_GROUPS; ++i) {
			var ret=groupCountBalance(i);
			myAPCount+=ret.ap;
			myATCount+=ret.at;
			myABCount+=ret.ab;
			myAACount+=ret.aa;
		}
		for (var i=0; i<NUM_GROUPS; ++i) {
			var ret=groupCountBalanceVtol(i);
			myAPCount+=ret.ap;
			myATCount+=ret.at;
			myABCount+=ret.ab;
			myAACount+=ret.aa;
		}
		myAACount+=groupSize(aaGroup);
		myAACount+=groupSize(aaVtolGroup);
		myABCount+=groupSize(abVtolGroup);
		// count stationary AA as well
		var list=enumStruct(me);
		for (var i=0; i<list.length; ++i)
			if (isDedicatedAA(list[i])) 
				myAACount+=.1; //Don't rely heavily on stationary anti-air.
		var ret=adaptationMatrix(
			myAPCount, myATCount, myAACount, myABCount, 
			enemyTankCount, enemyBorgCount, enemyVtolCount, enemyBuildingCount
		);
		RATE_AA = ret.aa;
		RATE_AB = ret.ab;
		RATE_AP = ret.ap;
		RATE_AT = ret.at;
	}
}

//Switch up rates at which tanks, cyborgs, and vtols are produced every now and then in an effort to become more unpredictable to enemies.
function changeProductionRates() {
	if (difficulty==EASY || difficulty == MEDIUM) //Only switch up production on HARD or INSANE difficulty.
		return;
	BUILD_TANKS = random(2);
	BUILD_BORGS = random(2);
	BUILD_VTOLS = random(2);
	
	//Make sure that the AI doesn't completely stop production of combat units.
	if (enumStruct(me,FACTORY).length && !BUILD_TANKS && (!enumStruct(me,CYBORG_FACTORY).length || !BUILD_BORGS) && (!enumStruct(me,VTOL_FACTORY).length || !BUILD_VTOLS))
		BUILD_TANKS = 1;
	else if (enumStruct(me,CYBORG_FACTORY).length && !BUILD_BORGS && (!enumStruct(me,FACTORY).length || !BUILD_TANKS) && (!enumStruct(me,VTOL_FACTORY).length || !BUILD_VTOLS))
		BUILD_BORGS = 1;
	else if (enumStruct(me,VTOL_FACTORY).length && !BUILD_VTOLS && (!enumStruct(me,FACTORY).length || !BUILD_TANKS) && (!enumStruct(me,CYBORG_FACTORY).length || !BUILD_BORGS))
		BUILD_VTOLS = 1;
	
	//If production rates for tanks, cyborgs, AND vtols are disabled, or if the AI finds itself in a life-threatening situation, then enable them all.
	if ((!BUILD_TANKS && !BUILD_BORGS && !BUILD_VTOLS) || status == STATUS_PANIC) {
		BUILD_TANKS = 1;
		BUILD_BORGS = 1;
		BUILD_VTOLS = 1;
	}
}

// switch to pure hovers as soon as at least one enemy base or factory is beyond the sea
function tankReachibilityCheck(droid) {
	if (!iHaveHovers()) // woundn't make much sense to switch to hovers when we don't have hovers
		return false;
	if (RATE_TANK==0) // we have already switched, no need to check further
		return false;
	for (var i=0; i<maxPlayers; ++i) if (isAnEnemy(i)) {
		if (!droidCanReach(droid,startPositions[i].x,startPositions[i].y))
			return true;
	}
	for (var i=0; i<maxPlayers; ++i) if (isAnEnemy(i)) {
		for (var j=0; j<allInAttackTargets.length; ++j) {
			var list = enumStruct(i,allInAttackTargets[j]);
			for (var k=0; k<list.length; ++k)
				if (!droidCanReach(droid,list[k].x,list[k].y))
					return true;
		}
	}
	return false;
}

//Create a new hotspot.
function createHotspot(x,y) {
	//Find the 1st available hotspot.
	for (var i=0; i<hotspots.length; ++i) {
		if (hotspots[i].lifespan<=0) {
			hotspots[i].lifespan=10;
			hotspots[i].x=x;
			hotspots[i].y=y;
			return;
		}
	}
	//If no available hotspot objects are found, then create a new one.
	hotspots[hotspots.length] = new constructHotspotInfo(x,y);
}

// returns a droid to the base
function forceReturnToBase(droid) {
	var list=enumStructArray(me,[lab,generator,]);
	if (list.length<=0)
		return;
	var i=random(list.length);
	orderDroidLoc(droid, DORDER_MOVE, list[i].x+1, list[i].y+1);
}

//Order trucks to demolish any partially-built structures near a hotspot.
function demolishHotspotStructs() {
	var structs=enumStruct(me);
	var trucks=enumGroup(oilerGroup);
	for (var j=0; j<structs.length; ++j) {
		if (structs[j].status==BEING_BUILT && structs[j].stattype!=RESOURCE_EXTRACTOR && structs[j].stattype!=WALL && structs[j].stattype!=GATE) {
			if (isNearHotspot(structs[j].x,structs[j].y) && !isNearBase(structs[j].x, structs[j].y)) {
				for (var k=0; k<trucks.length; ++k) {
					if (trucks[k].order!=DORDER_DEMOLISH && distBetweenTwoPoints(trucks[k].x, trucks[k].y, structs[j].x, structs[j].y) <= HOTSPOT_SIZE) {
						orderDroidObj(trucks[k],DORDER_DEMOLISH,structs[j]);
					}
				}
			}
		}
	}
}

// is the droid already at the base?
function droidIsAtHome(droid) {
	return isNearBase(droid.x, droid.y);
}

// returns a droid to the base if not at the base yet
function returnToBase(droid) {
	if (!droidIsAtHome(droid))
		forceReturnToBase(droid);
}

// returns list of all enemies, that is not re-generated too often
function cachedEnumEnemies(col) {
	if (typeof(cachedEnumEnemies.cache)=="undefined") 
		cachedEnumEnemies.cache=[];
	if (typeof(cachedEnumEnemies.renewTime)=="undefined") 
		cachedEnumEnemies.renewTime=[];
	if (typeof(col)=="undefined")
		col=-1;
	if (typeof(cachedEnumEnemies.cache[col])=="undefined" || (gameTime - cachedEnumEnemies.renewTime[col] > 5000)) {
		cachedEnumEnemies.cache[col] = [];
		for (var enemy=0; enemy<maxPlayers; ++enemy) if (col==-1 || col==enemy) if (isAnEnemy(enemy)) {
			cachedEnumEnemies.cache[col] = cachedEnumEnemies.cache[col].concat(enumDroid(enemy));
		}
		// can't run isVTOL() on cached units, so need to cache its value as well
		for (var i=0; i<cachedEnumEnemies.cache[col].length; ++i)
			if (isVTOL(cachedEnumEnemies.cache[col][i]))
				cachedEnumEnemies.cache[col][i].isVTOL=1;
			else
				cachedEnumEnemies.cache[col][i].isVTOL=0;
		for (var enemy=0; enemy<maxPlayers; ++enemy) if (col==-1 || col==enemy) if (isAnEnemy(enemy)) {
			cachedEnumEnemies.cache[col] = cachedEnumEnemies.cache[col].concat(enumStruct(enemy));
		}
		if (scavengerPlayer!=-1 && col==-1) {
			cachedEnumEnemies.cache[col] = cachedEnumEnemies.cache[col].concat(enumDroid(scavengerPlayer));
			cachedEnumEnemies.cache[col] = cachedEnumEnemies.cache[col].concat(enumStruct(scavengerPlayer));
		}
		cachedEnumEnemies.renewTime[col] = gameTime;
	}
	return cachedEnumEnemies.cache[col];
}

// estimates enemy activity at particular point
function dangerLevel(x,y) {
	if (DEF_LIGHT==1) {
		return 0;
	}
	if (difficulty == EASY || difficulty == MEDIUM)
		return (safeDest(me,x,y)?0:1); // a non-cheating procedure here
	const DIST=BASE_SIZE/2; // deity-cheating procedure for higher difficulties follows
	var list=[];
	var badGuys = 0;
	if (typeof(enumRange)!="undefined") {
		list = enumRange(x,y,DIST,ENEMIES,false);
	} else {
		list = cachedEnumEnemies();
	}
	for (var i=0; i<list.length; ++i) 
		if (distBetweenTwoPoints(list[i].x,list[i].y,x,y) <= DIST)
			if (list[i].type == DROID || (list[i].type==STRUCTURE && list[i].stattype==DEFENSE))
				++badGuys;
	return badGuys;
}

// returns a list of all repair facilities
function getRepairList() {
	var centers=[];
	for (var ally=0; ally<maxPlayers; ++ally) if (allianceExistsBetween(me,ally) || ally==me)
		for (var i=0; i<repair.length; ++i)
			centers = centers.concat(enumStruct(ally, repair[i]));
	return centers;
}

// checks if a droid needs to be sent for repair
function needsRepair(droid) {
	if (droid.health < RETREAT_AT) 
		return true;
	if (droid.health >= 99)
		return false;
	var centers = getRepairList();
	if (centers.length == 0)
		return false;
	for (var i=0; i<centers.length; ++i) {
		if (distBetweenTwoPoints(centers[i].x,centers[i].y,droid.x,droid.y) < 3)
			return true;
	}
	return false;
}

// sends a droid to a nearest repair center if necessary;
// returns true iff the droid was actually sent
function sendForRepair(droid) {
	if (status==STATUS_UNSTUCK)
		return false;
	if (!needsRepair(droid))
		return false;
	var centers = getRepairList();
	var gr=groupOfTank(droid);
	if (centers.length == 0)
		return false;
	var idx=-1, dist=Infinity;
	for (var i=0; i<centers.length; ++i) {
		if (centers[i].status != BUILT) 
			continue;
		var d=distBetweenTwoPoints(centers[i].x,centers[i].y,droid.x,droid.y);
		if (d<dist) {
			idx=i;
			dist=d;
		}
	}
	if (idx>-1)
		orderDroidLoc(droid,DORDER_MOVE,centers[idx].x,centers[idx].y);
	else
		returnToBase(droid);
	return true;
}

// checks if any droids around need to be sent for repair
function sendAllForRepair() {
	var list;
	if (typeof(sendAllForRepair.lastGroup)=="undefined")
		sendAllForRepair.lastGroup = 0;
	list=enumGroup(battleGroup[sendAllForRepair.lastGroup]);
	for (var i=0; i<list.length; ++i) 
		sendForRepair(list[i]);
	++sendAllForRepair.lastGroup;
	if (sendAllForRepair.lastGroup >= NUM_GROUPS)
		sendAllForRepair.lastGroup=0;
}

//Choose a team role to specialize in to assist your team
function workWithAllies() {
	/*if (!countAllies()) { 
		teamRole=TEAM_ROLE_NORMAL;
		return;
	}
	else*/
	{
		//We'll need to assign a group role based on our various team members' starting positions' distance from enemy starting positions.
		/*var teammates = [me,];
		for (var i=0; i<maxPlayers; ++i) {
			if (i!=me)
				if (isAlive(i))
					if (allianceExistsBetween(me,i))
						teammates[teammates.length]=i;
		}
		var minDist = [];
		var dist;
		var allyPos, enemyPos;
		for (i=0; i<teammates.length; ++i) {
			minDist[i] = Infinity;
			dist = Infinity;
			allyPos = startPositions[teammates[i]];
			for (var j=0; j<maxPlayers; ++j) {
				if (!isAlive(i))
					continue;
				if (j==teammates[i])
					continue;
				if (allianceExistsBetween(j,teammates[i]))
					continue;
				enemyPos = startPositions[j];
				dist = distBetweenTwoPoints(allyPos.x,allyPos.y,enemyPos.x,enemyPos.y);
				if (dist<minDist[i])
					minDist[i]=dist;
			}
		}
		//Now that we've got each teammate's distance to the closest enemy, let's see if we're closest to or furthest away from the front lines.
		var frontLine = true, rearLine = true;
		if (teammates.length>=2) {
			for (i=1; i<teammates.length; ++i) {
				if (minDist[0]>minDist[i])
					frontLine = false;
				if (minDist[0]<minDist[i])
					rearLine = false;
			}
		}
		else {
			frontLine = false;
			rearLine = false;
		}
		if (frontLine && !rearLine)
			teamRole=TEAM_ROLE_VANGUARD;
		else if (!frontLine && rearLine)
			teamRole=TEAM_ROLE_SUPPORT;
		else*/
			teamRole=TEAM_ROLE_NORMAL;
	}
	return;
}

//Our home base was sacked! Find a new place to rebuild and call home!
function findNewHomeBase() {
	//setStatus(STATUS_HOMELESS);
	//Look for any proxy factories.
	var list = [];
	var facList = enumStructArray(me,[FACTORY,CYBORG_FACTORY,VTOL_FACTORY,]);
	if (facList.length>0) {
		for (var i=0; i<facList.length; ++i) {
			if (!isNearBase(facList[i].x,facList[i].y))
				list[list.length]=facList[i];
		}
		if (list.length>0) {
			var r = random(list.length);
			basePosition = {x:list[r].x,y:list[r].y};
			return true;
		}
	}
	//If no proxy factories are found, then look for an ally's factories.
	var player=random(maxPlayers);
	for (var i=0; i<maxPlayers; ++i)
	{
		if (allianceExistsBetween(me,player) && player!=me && isAlive(player)) {
			list=enumStructArray(player,[FACTORY,CYBORG_FACTORY,VTOL_FACTORY,]);
			if (list.length>0) {
				var r = random(list.length);
				basePosition = {x:list[r].x,y:list[r].y};
				return true;
			}
		}
		player+=1;
		if (player>=maxPlayers)
			player=0;
	}
	//If no living allies are found, then look for our oil derricks.
	var oilList = enumStruct(me,RESOURCE_EXTRACTOR);
	if (oilList.length>0) {
		for (var i=0; i<oilList.length; ++i) {
			if (!isNearBase(oilList[i].x,oilList[i].y))
				list[list.length]=oilList[i];
		}
		if (list.length>0) {
			var r = random(list.length);
			basePosition = {x:list[r].x,y:list[r].y};
			return true;
		}
	}
	//If no suitable derricks are found, then look for any remaining trucks.
	var droidList = enumDroid(me,DROID_CONSTRUCT).concat(enumDroid(me,DROID_CYBORG_CONSTRUCT));
	if (droidList.length>0) {
		var r = random(droidList.length);
		basePosition = {x:droidList[r].x,y:droidList[r].y};
		return true;
	}
	return false;
}


/* Status manipulations */


// changes the status and adjusts necessary things
function setStatus(newStatus) {
	status=newStatus;
	switch(status) {
		case STATUS_NORMAL:
			statusTime=0;
			break;
		case STATUS_PANIC:
			statusTime=3;
			changeProductionRates(); //This will set all factories to produce units to defend the base.
			break;
		case STATUS_EARLYGAME:
			statusTime=30;
			break;
		case STATUS_UNSTUCK:
			statusTime=3;
			break;
		//case STATUS_HOMELESS:
			//statusTime=30;
			//break;
	}
}

// relaxes the status to normal and also panics sometimes
// called every 2 seconds
function updateStatus() {
	var goodGuys=0, badGuys=0;
	if (typeof(updateStatus.lastUpdate)=="undefined") {
		updateStatus.lastUpdate = -1;
	}
	var enemy=updateStatus.lastUpdate + 1;
	if (enemy>=maxPlayers)
		enemy=0;
	if (DEF_LIGHT==0) {
		var lastCount=enemyInfo[enemy].tanksAtBase;
		enemyInfo[enemy].tanksAtBase=0;
		var list = enumDroid(enemy,DROID_WEAPON,me);
		list = list.concat(enumDroid(enemy,DROID_CYBORG,me));
		var newCount=0;
		for (var i=0; i<list.length; ++i) 
			if (isNearBase(list[i].x,list[i].y)) {
				++enemyInfo[enemy].tanksAtBase;
			}
		for (var i=0; i<maxPlayers; ++i) {
			if (isAnEnemy(i))
				badGuys+=enemyInfo[i].tanksAtBase;
			else
				goodGuys+=enemyInfo[i].tanksAtBase;
		}
		if (scavengerPlayer!=-1) {
			var list=enumDroid(scavengerPlayer,DROID_WEAPON,me);
			list = list.concat(enumDroid(scavengerPlayer,DROID_PERSON,me),enumDroid(scavengerPlayer,DROID_TRANSPORTER,me));
			for (var i=0; i<list.length; ++i) {
				if (isNearBase(list[i].x,list[i].y)) {
					++badGuys;
					if (list[i].droidType==DROID_TRANSPORTER)
						attackedByChoppers=1;
				}
			}
		}
	}
	updateStatus.lastUpdate=enemy;
	//Relax need to build AA to counter Scavenger Choppers
	if (attackedByChoppers>0)
		attackedByChoppers-=0.2;
	if (badGuys>goodGuys) {
		if (status!=STATUS_PANIC) {
			callForHelp();
 		}
		setStatus(STATUS_PANIC);
		return;
	}
	if (statusTime>0)
		statusTime-=0.2;
	else {
		if (typeof(chat)!="undefined") {
			if (status==STATUS_PANIC) {
				chat(ALLIES,"Should be fine now!");
				removeBeacon(ALLIES);
			}
		}
		setStatus(STATUS_NORMAL);
	}
	//Gradually make hotspots safe again.
	for (var i=0; i<hotspots.length; ++i) {
		hotspots[i].lifespan-=2;
		if (hotspots[i].lifespan<0)
			hotspots[i].lifespan=0;
	}
}


/* Constructing the base */


// make sure builderGroup is non-empty, 
// make sure oiler group is non-empty, 
// remix underpacked combat groups
function balanceGroups() {
	// HACK: this next check contains an ugly api level check to see whether we can produce cyborg engineers
	var iHaveFactory = iHaveStruct(factory) || (iHaveStruct(borgfac) && typeof(DORDER_RTB)!="undefined");
	// this fires if we have no factory to produce trucks and no builderGroup trucks to build a factory
	if (groupSize(builderGroup)==0 && groupSize(oilerGroup)>0 && !iHaveFactory) {
		var trucks=enumGroup(oilerGroup);
		for (var i=0; i<trucks.length; ++i) {
			groupAddDroid(builderGroup,trucks[i]);
			return; // add only one truck
		}
	}
	// this fires when we have no oilerGroup trucks to capture oil and no oil to produce oilerGroup trucks
	if (groupSize(oilerGroup)==0 && groupSize(builderGroup)>0 && iHaveFactory && iHaveStruct(generator) && !iHaveStruct(derrick)) {
		var trucks=enumGroup(builderGroup);
		for (var i=0; i<trucks.length; ++i) {
			groupAddDroid(oilerGroup,trucks[i]);
			return; // add only one truck
		}
	}
	// this fires when we have too many groups that don't have enough units inside
	var size=[];
	for (var i=0; i<NUM_GROUPS; ++i)
		size[i]=groupSize(battleGroup[i]);
	for (var i=0; i<NUM_GROUPS; ++i) if (size[i]<MIN_WARRIORS && size[i]>0) {
		for (var j=0; j<NUM_GROUPS; ++j) if (i!=j && size[j]<MIN_WARRIORS && size[j]>0) {
			if (size[j]>size[i]) { // merge a smaller group into a bigger group
				var k=j; j=i; i=k;
			}
			var list=enumGroup(battleGroup[j]);
			for (var k=0; k<list.length; ++k) {
				groupAddDroid(battleGroup[i],list[k]);
			}
			return; // handle only one pair of groups at a time
		}
	}
	for (var i=0; i<NUM_VTOL_GROUPS; ++i)
		size[i]=groupSize(vtolGroup[i]);
	for (var i=0; i<NUM_VTOL_GROUPS; ++i) if (size[i]<MIN_VTOLS && size[i]>0) {
		for (var j=0; j<NUM_VTOL_GROUPS; ++j) if (i!=j && size[j]<MIN_VTOLS && size[j]>0) {
			if (size[j]>size[i]) { // merge a smaller group into a bigger group
				var k=j; j=i; i=k;
			}			
			var list=enumGroup(vtolGroup[j]);
			for (var k=0; k<list.length; ++k) {
				groupAddDroid(vtolGroup[i],list[k]);
			}
			return; // handle only one pair of groups at a time
		}
	}
}


// checks if a truck is busy doing something
function isTruckBusy(truck) {
	if (truck.order == DORDER_BUILD)
		return true;
	if (truck.order == DORDER_HELPBUILD)
		return true;
	if (truck.order == DORDER_LINEBUILD)
		return true;
	if (truck.order == DORDER_DEMOLISH)
		return true;
	if (typeof(DORDER_RECYCLE)!="undefined")
		if (truck.order == DORDER_RECYCLE)
			return true;
	return false;
}

// order the builderGroup to build some base structure somewhere
function buildBasicStructure(struct,forceLevel,maxBlockingTiles,forceRebuild) {
	if (typeof(buildBasicStructure.lastCall)=="undefined")
		buildBasicStructure.lastCall=gameTime;
	if (gameTime - buildBasicStructure.lastCall < 2000)
		return false;
	if (typeof(forceLevel)=="undefined")
		var forceLevel=1;
	if (typeof(maxBlockingTiles)=="undefined")
		var maxBlockingTiles=0;
	if (typeof(forceRebuild)=="undefined") //forceRebuild helps keep builderGroup from having to move around cliffs to build the structure.
		var forceRebuild=0;
	if (typeof(struct)=="undefined")
		return false;
	var droidlist = enumGroup(builderGroup);
	if (droidlist.length == 0) 
		return false;
	var found=false;
	for (var i=0; i<droidlist.length; ++i)
		if (!isTruckBusy(droidlist[i]) || forceRebuild) {
			if (!found) {
				loc=pickStructLocation(
					droidlist[i], 
					struct, 
					basePosition.x, 
					basePosition.y, 
					maxBlockingTiles
				);
				if (typeof(loc)=="undefined" || earlyGame(personality.PEACE_TIME) || forceRebuild) {
					loc=pickStructLocation(
						droidlist[i], 
						struct, 
						droidlist[i].x, 
						droidlist[i].y, 
						maxBlockingTiles
					);
				}
				if (typeof(loc)=="undefined")
					return false;
				if (dangerLevel(loc.x,loc.y)>0 && forceLevel == 0)
					return false;
				found=true;
				if (struct == command) // where is our base?
					basePosition=loc;
				//basicBuildingQueue=struct; //We need to make sure that the trucks won't have to move around cliffs to build the new structure.
				//basicBuildingLoc=loc;
				//queue("basicBuildingCheckup",2000);
			}
			orderDroidBuild(droidlist[i], DORDER_BUILD, struct, loc.x, loc.y);
			buildBasicStructure.lastCall=gameTime;
		}
	return found;
}

function basicBuildingCheckup() {
	//const DACTION_BUILD=2, DACTION_BUILD_FOUNDATION=3; //Hack until droid actions are exposed to JavaScript API.
	var droidList = enumGroup(builderGroup);
	var success=false;
	for (var i=0; i<droidList; ++i) { //Have any of our builderGroup trucks made it to their destination yet?
		if (distBetweenTwoPoints(droidList[i].x,droidList[i].y,basicBuildingLoc.x,basicBuildingLoc.y)<8)
			success=true;
	}
	if (!success)
		buildBasicStructure(basicBuildingQueue,1,1,1);
}

// demolish some hopefully currently unnecessary structure
function recycleStructure(struct) {
	var droidlist = enumGroup(builderGroup);
	if (droidlist.length == 0)
		return false;
	for (var i=0; i<droidlist.length; ++i) if (!isTruckBusy(droidlist[i])) {
		orderDroidObj(droidlist[i], DORDER_DEMOLISH, struct);
		return true;
	}
	return false;
}

// build a specific module
function buildModule(struct, moduleType) {
	if (typeof(struct)=="undefined")
		return false;
	if (!isStructureAvailable(moduleType, me))
		return false;
	//Don't send base trucks to build modules for proxy factories.
	if (isBaseStructure(struct) || struct.stattype==RESEARCH_LAB || struct.stattype==POWER_GEN) {
		var droidlist = enumGroup(builderGroup);
		var success = false;
		for (var i=0; i<droidlist.length; ++i)	
			if (!isTruckBusy(droidlist[i]))
				if (orderDroidBuild(droidlist[i], DORDER_BUILD, moduleType, struct.x, struct.y))
					success = true;
		if (success)
			return true;
	}
	else { //Use one of our oiler group trucks to upgrade proxy factories.
		var droidlist = sortByDistance(struct.x,struct.y,enumGroup(oilerGroup));
		for (var i=0; i<droidlist.length; ++i)	
			if (!isTruckBusy(droidlist[i]))
				if (orderDroidBuild(droidlist[i], DORDER_BUILD, moduleType, struct.x, struct.y))
					return true;
	}
	return false;
}

// build counter-battery towers
function buildCB() {
	var x=lastArtyAttackX, y = lastArtyAttackY;
	if (x == -1 || y == -1)
		return false;
	if (distanceToNearestStructure(x,y,cbtower)<MIN_SENSOR_DISTANCE/2.5) {
		lastArtyAttackX = lastArtyAttackY = -1;
		return false;
	}
	if (buildTower(x,y,cbtower,1,0)) {
		lastArtyAttackX = lastArtyAttackY = -1;
		return true;
	}
	return false;	
}

// build radar detector towers
function buildDetector() {
	var x=lastArtyAttackX, y = lastArtyAttackY;
	if (x == -1 || y == -1)
		return false;
	if (distanceToNearestStructure(x,y,detector)<MIN_SENSOR_DISTANCE/2.5) {
		lastArtyAttackX = lastArtyAttackY = -1;
		return false;
	}
	if (buildTower(x,y,detector,1,0)) {
		lastArtyAttackX = lastArtyAttackY = -1;
		return true;
	}
	return false;	
}

// build anti-air defenses at specific spots
function buildAA() {
	var x=lastAirAttackX, y = lastAirAttackY;
	if (x == -1 || y == -1)
		return false;
	var i=random(personality.antiair.length);
	if (buildTower(x,y,personality.antiair[i],1,0)) {
		lastAirAttackX = lastAirAttackY = -1;
		return true;
	}
	return false;	
}

// build defensive structures on gateways (v3.2+ only)
function buildGateways() {
	// what to do on v3.1, when the gateway list is not exposed to scripts yet
	if (typeof(enumGateways)=="undefined" || DEF_LIGHT == 1) {
		var i=random(personality.hardpoints.length);
		return buildTower(basePosition.x+random(10)-5,basePosition.y+random(10)-5,personality.hardpoints[i],1,0);
	}
	if (personality.hardpoints.length == 0) 
		return false;
	var gateways = enumGateways();
	if (gateways.length == 0)
		return false;
	var gate=gateways[random(gateways.length)];
	if (getNearestCCPlayer(gate.x2,gate.y2,1) != me) {
		queue("buildGateways",100);
		return;
	}
	if (gate.y2 < gate.y1) {
		var t=gate.y1;
		gate.y1=gate.y2;
		gate.y2=t;
	}
	if (gate.x2 < gate.x1) {
		var t=gate.x1;
		gate.x1=gate.x2;
		gate.x2=t;
	}
	var i=random(personality.hardpoints.length);
	return buildTower(gate.x1+random(gate.x2-gate.x1+1),gate.y1+random(gate.y2-gate.y1+1),personality.hardpoints[i],1,0);
}

// how many defenses do i have?
function countMyDefenses() {
	var list=enumStruct(me,DEFENSE);
	var count=0;
	for (var i=0; i<list.length; ++i)
		if (!isDedicatedAA(list[i]) && !isLongRangeArtillery(list[i])) //Exclude anti-air and long range artillery.
			++count;
	return count;
}

// build some defensive structure somewhere near (x,y)
function buildTower(x,y,list,maxDangerLevel,forceLevel,maxBlockingTiles) {
	if (typeof(maxBlockingTiles)=="undefined")
		var maxBlockingTiles=0;
	if (x<2) x=2;
	if (y<2) y=2;
	if (x>mapWidth-3) x=mapWidth-3;
	if (y>mapHeight-3) y=mapHeight-3;
	if (maxDangerLevel == 0 && !safeDest(me,x,y))
		return false;
	if (isNearHotspot(x,y))
		return false;
	var droidlist = sortByDistance(x,y,enumGroup(oilerGroup));
	if (droidlist.length == 0 || list.length == 0) 
		return false;
	var tower;
	for (var i=0; i<list.length; ++i)
		if (isStructureAvailable(list[i],me)) {
			tower=list[i];
			break;
		}
	if (typeof(tower)=="undefined")
		return false;
	var loc = pickStructLocation(droidlist[0], tower, x, y, maxBlockingTiles);
	if (typeof(loc)=="undefined")
		return false;
	if (list==sensors) { // not really the best way to check ...
		if (distanceToNearestStructure(loc.x,loc.y,sensors)<MIN_SENSOR_DISTANCE && distanceToNearestStructure(loc.x,loc.y,[command,])<MIN_SENSOR_DISTANCE) {
			return false;
		}
	}
	if (maxDangerLevel == 0 && !safeDest(me,loc.x,loc.y))
		return false;
	if (isNearHotspot(loc.x,loc.y))
		return false;
	for (var j=0; j<forceLevel+1; ++j) { //We should check for trucks that aren't busy before forcing already busy trucks to abandon their current tasks.
		for (var i=0; i<droidlist.length; ++i) {
			if (j>0 || (!isTruckBusy(droidlist[i]) && droidlist[i].order != DORDER_MOVE)) {
				/*if (personality.DEFENSIVENESS>0) 
					if (forceLevel==0 && random(personality.DEFENSIVENESS)>0)
						return false;*/
				orderDroidBuild(droidlist[i], DORDER_BUILD, tower, loc.x, loc.y); 
				return true;
			}
		}
	}
	return false;
}

// demolish outdated towers
function sellOldDefenses() {
	var doublelist=personality.defenses;
	doublelist=doublelist.concat(personality.hardpoints);
	for (var i=0; i<doublelist.length; ++i)
		for (var j=3; j<doublelist[i].length; ++j)
			if (isStructureAvailable(doublelist[i][j-3],me)) {
				var list=enumStruct(me,doublelist[i][j]);
				if (list.length>0) {
					recycleStructure(list[0]);
					return;
				}
			}
	
}

// returns the command center owner nearest to (x,y)
function getNearestCCPlayer(x,y,all) {
	var minDist=Infinity, minPlayer;
	for (var ally=0; ally<maxPlayers; ++ally) if (allianceExistsBetween(me,ally) || me==ally || all==1) {
		var list=enumStruct(ally,command);
		for (var i=0; i<list.length; ++i) {
			var d=distBetweenTwoPoints(list[i].x,list[i].y,x,y);
			if (d<minDist) {
				minDist=d;
				minPlayer=ally;
			}
		}
	}
	return minPlayer;
}


//A more advanced version of enumStruct() that returns an array of structures within the specified array.
function enumStructArray(player,structList,viewer) {
	//Uses the same optional parameters, and their defaults, as enumStruct.
	if (typeof(player) == "undefined")
		var player = me;
	if (typeof(viewer) == "undefined")
		var viewer = -1;
	var list=[];
	for (var i=0; i<structList.length; ++i)
		list=list.concat(enumStruct(player,structList[i],viewer));
	return list;
}

// finds out the distance to the nearest structure of a certain type
/*function distanceToNearestStructure(x,y,structurelist,player) {
	if (typeof(player) == "undefined")
		var player = me;
	var minDist=Infinity;
	for (var i=0; i<structurelist.length; ++i) {
		var list=enumStruct(player,structurelist[i]);
		for (var j=0; j<list.length; ++j) {
			var dist=distBetweenTwoPoints(x,y,list[j].x,list[j].y);
			if (dist<minDist)
				minDist=dist;
		}
	}
	return minDist;
}*/

// finds out the distance to the nearest structure of a certain type
function distanceToNearestStructure(x,y,structurelist,player) {
	if (typeof(player) == "undefined")
		var player = me;
	var list=[];
	for (var i=0; i<structurelist.length; ++i)
		list=list.concat(enumStruct(player,structurelist[i]));
	if (list.length==0)
		return Infinity;
	var struct=findNearestObject(x,y,list);
	var minDist=distBetweenTwoPoints(x,y,struct.x,struct.y);
	return minDist;
}

//Build factories in remote areas
function buildProxyFactory() {
	if ((earlyGame(personality.PEACE_TIME) && RATE_TANK>0) || difficulty==EASY || difficulty==MEDIUM)
		return false; //EASY and MEDIUM AIs shouldn't think about building factories outside of their bases.
	var oil = findMostValuableDerrick();
	if (typeof(oil)=="undefined") {
		return false; //We should already be building factories within our base thanks to the keepBuildingThings() function.
	}
	if (distanceToNearestStructure(oil.x,oil.y,[factory,borgfac,vtolfac,]) < BASE_SIZE)
		return false;
	var faclist=[[]];
	if (personality.THIS_AI_MAKES_TANKS) {
		faclist[faclist.length]=[factory,];
		faclist[faclist.length]=[factory,];
		faclist[faclist.length]=[factory,];
	}
	if (personality.THIS_AI_MAKES_CYBORGS) {
		faclist[faclist.length]=[borgfac,];
		faclist[faclist.length]=[borgfac,];
		faclist[faclist.length]=[borgfac,];
	}
	if (personality.THIS_AI_MAKES_VTOLS) {
		faclist[faclist.length]=[vtolfac,];
	}
	if (faclist.length==0)
		return false;
	//Limit construction of new factories to 1 per 4 derricks on low power settings on HARD or easier difficulty.
	var factoryCount = enumStructArray(me,[factory,borgfac,]).length;
	var genCount = enumStruct(me, generator).length;
	var derrickCount = enumStruct(me, derrick).length;
	var powerMultiplier = powerType*(1+(difficulty==INSANE))/400;
	if (genCount<factoryCount/(powerMultiplier) || derrickCount<4*factoryCount/(powerMultiplier))
		return false;
	var success = buildTower(oil.x,oil.y,faclist[random(faclist.length)],0,0,1);
	//if (success)
		//queue("buildRepairFacility",300);
	return success;
}

//Build repair facilities in remote areas
function buildRepairFacility() {
	if (earlyGame(personality.PEACE_TIME) || difficulty==EASY || (!personality.THIS_AI_MAKES_TANKS && !personality.THIS_AI_MAKES_CYBORGS))
		return false; //EASY AIs don't retreat damaged units, so they shouldn't find any use for repair facilities.
	var oil = findMostValuableDerrick();
	if (typeof(oil)=="undefined") {
		return false; //We should already be building a Repair Facility within our base thanks to the keepBuildingThings() function.
	}
	//Limit construction of new repair facilities to 1 per factory.
	var factoryCount=enumStructArray(me,[factory,borgfac,]).length;
	var repairCount = enumStruct(me, REPAIR_FACILITY).length;
	if (repairCount>=factoryCount)
		return false;
	if (distanceToNearestStructure(oil.x,oil.y,repair) < BASE_SIZE)
		return false;
	var success = buildTower(oil.x,oil.y,repair,0,0,1);
	//if (success)
		//queue("buildRepairFacility",300);
	return success;
}

//Build artillery structures in remote areas
function buildArtillery() {
	if (earlyGame(personality.PEACE_TIME) || !personality.THIS_AI_BUILDS_ARTILLERY) //EASY AIs shouldn't understand how sensors and artillery work.
		return false;
	var maxBlockingTiles=0;
	if (random(2)>=0) {
		var target=findTarget(-1);
		if (typeof(target)!="undefined") {
			var list=enumStruct(me,derrick);
			for (var i=0; i<maxPlayers; ++i)
				if (allianceExistsBetween(me,i))
					list=list.concat(enumStruct(i,derrick)); //List allied derricks as well.
			list=sortByDistance(target.x,target.y,list);
			if (list.length) {
				var j=list.length
				for (var i=0; i<j; ++i)
					for (var k=0; k<(j-i); ++k)
						list=list.concat(list[i]); //Give greater priority to the derricks that are closer to the front lines.
				var oil=list[random(list.length)];
			}
		}
	}
	/*else if (random(2)==0)
		var oil=findHottestDerrick();
	else
		var oil=findMostValuableDerrick();*/
	if (typeof(oil)=="undefined") {
		oil={x:basePosition.x,y:basePosition.y};
		//maxBlockingTiles=0;
	}
	//We need to build and maintain a certain amount of artillery structures,
	//even if it means pulling some of our oilerGroup trucks away from whatever they were building.
	var artCount=0;
	var list=enumStruct(me,DEFENSE);
	for (var i=0; i<list.length; ++i)
		if (isLongRangeArtillery(list[i]))
			artCount+=1;
	if (artCount<MIN_ARTILLERY)
		var forceLevel=false;
	else 
		var forceLevel=true;
	if (random(2)==0 && !forceLevel) {
		var r=random(3);
		for (var i=0; i<3; ++i) {
			switch (r) {
			case 0:
				if (distanceToNearestStructure(oil.x,oil.y,sensors) > 26 && distanceToNearestStructure(oil.x,oil.y,[command,]) > 26)
					if (buildTower(oil.x,oil.y,sensors,1,0,maxBlockingTiles))
						return;
			case 1:
				if (distanceToNearestStructure(oil.x,oil.y,cbtower) > 26)
					if (buildTower(oil.x,oil.y,cbtower,1,0,maxBlockingTiles))
						return;
			case 2:
				if (distanceToNearestStructure(oil.x,oil.y,detector) > 26)
					if (buildTower(oil.x,oil.y,detector,1,0,maxBlockingTiles))
						return;
			}
			r+=1;
			if (r>=3)
				r-=3;
		}
	}
	if (!iHaveStruct(command) && !personality.fixedTemplates) 
		return;
	var art = personality.artillery;
	if (random(RATE_AP+RATE_AT+RATE_AB)<RATE_AP)
		art = personality.apArtillery;
	else if (random(RATE_AT+RATE_AB)<RATE_AT)
		art = personality.atArtillery;
	else
		art = personality.abArtillery;
	var success = buildTower(oil.x,oil.y,art[random(art.length)],0,forceLevel,maxBlockingTiles);
	//if (success)
		//queue("buildArtillery",300);
	return success;
}

//Build machinegun guard towers close to unclaimed oil resources and enemy oil derricks.
function buildLookoutTower() {
	if (!isStructureAvailable("GuardTower1",me))
		return false;
	var list=enumFeature(-1,oilres);
	var success=false;
	for (var i=0; i<11; ++i)
		if ((isAnEnemy(i) || i==scavengerPlayer) && i!=me)
			list=list.concat(enumStruct(i,derrick));
	if (list.length=0)
		return false;
	list=sortByDistance(basePosition.x,basePosition.y,list);
	var k=list.length;
	for (i=0; i<k; ++i)
		for (var j=0; j<(k-i); ++j)
			list=list.concat(list[i]); //Give extra priority to closer resources and enemy derricks.
	var pos=list[random(list.length)];
	if (typeof(pos)=="undefined")
		return false;
	if (distanceToNearestStructure(pos.x,pos.y,["GuardTower1",],me)>6)
		success = buildTower(pos.x,pos.y,["GuardTower1",],2,0,1);
	//if (success)
		//queue("buildLookoutTower",300);
	return success;	
}

//Build anti-air defenses close to our oil derricks.
function buildAADefense() {
	if (random(MIN_DEFENDERS_AA+1)<1)
		return false;
	var success=false;
	var list=enumStruct(me,derrick);
	for (var i=0; i<maxPlayers; ++i)
		if (allianceExistsBetween(me,i))
			list=list.concat(enumStruct(i,derrick)); //List allied derricks as well.
	if (list.length=0)
		return false;
	var pos=list[random(list.length)];
	if (typeof(pos)=="undefined")
		return false;
	var aaDefenses = [];
	for (var i=0; i<personality.antiair.length; ++i)
		aaDefenses=aaDefenses.concat(personality.antiair[i]);
	if (distanceToNearestStructure(pos.x,pos.y,aaDefenses,me)>6)
		success = buildTower(pos.x,pos.y,personality.antiair[random(personality.antiair.length)],1,0,1);
	//if (success)
		//queue("buildAADefense",300);
	return success;	
}

//Build the occasional Wind Farm
function buildWindFarm() {
	if (status!=STATUS_NORMAL || !isStructureAvailable(windfarm,me))
		return false;
	if (buildBasicStructure(windfarm,1,0));
		return true;
	return buildTower(basePosition.x,basePosition.y,[windfarm,],1,0,0); //This is in case all builderGroup trucks are currently tied up building other things.
}

// protect things
function buildDefenses() {
	if (earlyGame(personality.PEACE_TIME))
		return false;
	if (!iHaveStruct(command) && !personality.fixedTemplates) 
		return;
	var success=false;
	var oil=findHottestDerrick();
	var maxBlockingTiles=1;
	if (typeof(oil)=="undefined") {
		oil={x:basePosition.x,y:basePosition.y};
		maxBlockingTiles=0;
	}
	if (personality.DEFENSIVENESS == 0)
		queue("buildGateways",100);
	if (RATE_AA>0 && (difficulty==HARD || difficulty == INSANE)) {
		success=buildTower(basePosition.x,basePosition.y,personality.antiair[random(personality.antiair.length)],0,0);
	} 
	var myDefenses=countMyDefenses();
	var tooManyDefenses=(personality.DEFENSIVENESS!=0 && myDefenses>30/Math.pow(personality.DEFENSIVENESS,0.5));
	if (personality.DEFENSIVENESS==0)
		tooManyDefenses=(random(100)<myDefenses);
	if (!success) {
		switch(random(4)) {
			/*case 4:
				success=buildTower(oil.x,oil.y,sensors,2,0,maxBlockingTiles);
				if (success) break;*/
			case 3:
			case 2:
				if (personality.THIS_AI_BUILDS_ARTILLERY) { //EASY AIs shouldn't understand how to use artillery effectively, and should thus not bother using them.
					var art = personality.artillery;
					if (random(RATE_AP+RATE_AT+RATE_AB)<RATE_AP)
						art = personality.apArtillery;
					else if (random(RATE_AT+RATE_AB)<RATE_AT)
						art = personality.atArtillery;
					else
						art = personality.abArtillery;
					success=buildTower(basePosition.x+random((BASE_SIZE*2)+1)-BASE_SIZE,basePosition.y+random((BASE_SIZE*2)+1)-BASE_SIZE,art[random(art.length)],0,0);
				}
				if (success) break;
			case 1:
				if (tooManyDefenses) {
					sellOldDefenses(); 
					break;
				}
				success=buildGateways();
				if (success) break;
			case 0:
				if (tooManyDefenses) {
					sellOldDefenses(); 
					break;
				}
				var i=random(personality.defenses.length);
				success=buildTower(oil.x,oil.y,personality.defenses[i],0,0,maxBlockingTiles);
				if (success) break;
		}
	}
	if (success)
		queue("buildDefenses",300);
}

// expand the base by building more labs and factories
function keepBuildingThings() {
	if (earlyGame(personality.PEACE_TIME) && difficulty != EASY && RATE_TANK>0) //Don't do this at the start of the game (unless playing on a hover-only map)!
		return;
	var labCount = enumStruct(me, lab).length;
	var factoryCount = enumStruct(me, factory).length;
	var borgfacCount = enumStruct(me, borgfac).length;
	var vtolfacCount = enumStruct(me, vtolfac).length;
	var ccCount = enumStruct(me, command).length;
	var genCount = enumStruct(me, generator).length;
	var repairCount = enumStruct(me, repair).length;
	var cbCount = enumStruct(me, cbtower).length;
	var detectorCount = enumStruct(me, detector).length;
	var derrickCount = enumStruct(me, derrick).length;

	var list = [];

	//Count our anti-air defenses as well.
	/*var aaCount = 0;
	var defenseList = enumStruct(me, DEFENSE);
	for (var i=0; i<defenseList.length; ++i) {
		if (isNearBase(defenseList[i].x,defenseList[i].y)) {
			if (isDedicatedAA(defenseList[i]))
				aaCount+=1;
			else
				for (var j=0; j<personality.antiair.length; ++j)
					for (var k=0; k<personality.antiair[j].length; ++k)
						if (defenseList[i].name==personality.antiair[j][k])
							aaCount+=1;
		}
	}*/
	// force building at least something when build order fails due to danger
	if (genCount < 1 && isStructureAvailable(generator, me))
		return buildBasicStructure(generator,1,1);
	if (factoryCount < 1 && isStructureAvailable(factory, me))
		return buildBasicStructure(factory,1,1);
	if (ccCount < 1 && isStructureAvailable(command, me))
		return buildBasicStructure(command,1,1);
	if (labCount < 1 && isStructureAvailable(lab, me))
		return buildBasicStructure(lab,1,1);
	if (borgfacCount < 1 && isStructureAvailable(borgfac, me) && personality.THIS_AI_MAKES_CYBORGS==true && RATE_TANK>0)
		return buildBasicStructure(borgfac,1,1);
	if (vtolfacCount < 1 && isStructureAvailable(vtolfac, me) && personality.THIS_AI_MAKES_VTOLS==true)
		return buildBasicStructure(vtolfac,1,1);
	// build generators when necessary
	if (genCount < 2 || derrickCount >= 4*genCount) {
		if (isStructureAvailable(generator,me)) {
			if (UNUSUAL_SITUATION==0) {
				if (playerPower(me) < LOW_POWER) {
					buildBasicStructure(generator,0,1);
					return;
				}
			}
		}
	}
	// some advanced build order stuff
	if (difficulty!=EASY) { //EASY AIs should not understand the concepts of repairing damaged droids.
		if (isStructureAvailable(repair,me) && (repairCount < 1 || distanceToNearestStructure(basePosition.x,basePosition.y,repair) > BASE_SIZE)) 
			for (i=0; i<repair.length; ++i) //Build a repair facility at our home base.
				if (buildBasicStructure(repair[i],1,1))
					return true;
	}
	if (personality.THIS_AI_BUILDS_ARTILLERY) { //EASY AIs should not understand how to use artillery.
		if (isStructureAvailable(cbtower,me) && (cbCount < 1 || distanceToNearestStructure(basePosition.x,basePosition.y,cbtower) > BASE_SIZE)) 
			for (i=0; i<cbtower.length; ++i) //Build a counter-battery tower at our home base.
				if (buildBasicStructure(cbtower[i],1,1))
					return true;
		else if (isStructureAvailable(detector,me) && (detectorCount < 1 || distanceToNearestStructure(basePosition.x,basePosition.y,detector) > BASE_SIZE)) 
			for (i=0; i<repair.detector; ++i) //Build a radar detector tower at our home base.
				if (buildBasicStructure(detector[i],1,1))
					return true;
	}
	if (isStructureAvailable(lassat,me) && enumStruct(me,lassat).length <1)
		return buildBasicStructure(lassat,0,1);
	if (isStructureAvailable(satlink,me) && enumStruct(me,satlink).length <1)
		return buildBasicStructure(satlink,0,1);

	//Build a lab for every 4 derricks on low power settings on HARD or easier difficulty.
	//Build a lab for every 2 derricks on low power settings on INSANE difficulty.
	//(Should build a lab for every 1.6 derricks on high power settings, or for every .8 derricks on INSANE difficulty.)
	var powerMultiplier = powerType*(1+(difficulty==INSANE))/400;
	if (genCount>=(labCount/powerMultiplier) && derrickCount>=(4*labCount/powerMultiplier) && isStructureAvailable(lab, me))
		return buildBasicStructure(lab,1,1);
	//Build anti-air defenses when necessary.
	/*if (aaCount<MIN_DEFENSES_AA) {
		list = personality.antiair[random(personality.antiair.length)];
		for (i=0; i<list.length; ++i)
			if (isStructureAvailable(list[i],me))
				if (buildBasicStructure(list[i]))
					return true;
	}*/
	// build VTOL pads when necessary
	if (personality.THIS_AI_MAKES_VTOLS) {
		if (enumStruct(me, vtolfac).length > 0) {
			var vtolList = enumDroid(me, DROID_WEAPON);
			var vtolCount=0;
			for (var i=0; i<vtolList.length; ++i)
				if (isVTOL(vtolList[i]))
					++vtolCount;
			var padCount = enumStruct(me, vtolpad).length;
			if ((NUM_VTOL_PADS*padCount < vtolCount || (padCount < vtolCount && playerPower(me) > EXTREME_LOW_POWER)) && isStructureAvailable(vtolpad,me)) {
				buildBasicStructure(vtolpad,0);
				return;
			}
		}
	}
	if (isStructureAvailable(lassat,me)) //Build additional LasSat Command Posts if the game will allow it.
		return buildBasicStructure(lassat,0,1);
	//if (RATE_TANK==0 && isStructureAvailable(lab,me) && enumStruct(me,lab).length<4 && !iHaveHovers())
	if ((genCount/2)>=(labCount/powerMultiplier) && (derrickCount/2)>=(4*labCount/powerMultiplier) && isStructureAvailable(lab, me) && RATE_TANK==0 && !iHaveHovers())
		return buildBasicStructure(lab,1,1); // research hovers faster when having no ground enemies
	if (earlyGame(personality.PEACE_TIME) && RATE_TANK==0 && !iHaveHovers())
		return buildWindFarm();
	else {
		//Keep building research labs on EASY difficulty. (EASY AIs shouldn't know better than to do this if low on power.)
		if (difficulty==EASY && isStructureAvailable(lab, me))
			return buildBasicStructure(lab,1,1);
		//build some long-range artillery, repair facilities, and proxy factories if available
		queue("buildArtillery");
		if (random(3)==0)
			queue("buildProxyFactory");
		if (random(2)==0)
			queue("buildRepairFacility");
		// build defenses
		if (personality.DEFENSIVENESS!=0 && random(personality.DEFENSIVENESS+1) != 0 && playerPower(me) > LOW_POWER)
			queue("buildDefenses");
		// use extra power for expansion
		if (playerPower(me) > 2000) {
			queue("buildRepairFacility");
			queue("buildProxyFactory");
			if (isStructureAvailable(lab, me) && !researchBlocked())
				return buildBasicStructure(lab,1,1);
			if (difficulty<=MEDIUM && random(4)==0) {
				var list=[];
				if (isStructureAvailable(factory,me) && personality.THIS_AI_MAKES_TANKS)
					list[list.length]=factory;
				if (isStructureAvailable(borgfac,me) && personality.THIS_AI_MAKES_CYBORGS && RATE_TANK>0)
					list[list.length]=borgfac;
				if (isStructureAvailable(vtolfac,me) && personality.THIS_AI_MAKES_VTOLS)
					list[list.length]=vtolfac;
				buildBasicStructure(list[random(list.length)],0,1);
			}
			//Spend remaining power building as much artillery as possible.
			if (personality.THIS_AI_BUILDS_ARTILLERY) {
				var art = personality.artillery;
				if (random(RATE_AP+RATE_AT+RATE_AB)<RATE_AP)
					art = personality.apArtillery;
				else if (random(RATE_AT+RATE_AB)<RATE_AT)
					art = personality.atArtillery;
				else
					art = personality.abArtillery;
				return buildTower(basePosition.x+random((BASE_SIZE*2)+1)-BASE_SIZE,basePosition.y+random((BASE_SIZE*2)+1)-BASE_SIZE,art[random(art.length)],0,0); 
			}
		}
	}	
}

// the main function responsible for building things
// calls other functions in a particular order
function executeBuildOrder() {
	if (!finishBuildings())
		if (!myBuildOrder())
			if (!researchRush())
				if (!secureBaseOil())
					if (!buildAdditionalGenerators())
						if (!upgradeGenerators())
							if (!upgradeStructures())
								if (!outOfPower()) //Don't build things if there isn't any power left.
									//if (!buildCB())
										//if (!buildDetector())
											//if (!buildAA())
												queue("keepBuildingThings");
}

// should we panic if this structure is under attack or unfinished?
function isBaseStructure(struct) {
	return (isNearBase(struct.x,struct.y));
}

// finish buildings that were never finished yet
function finishBuildings() {
	var droidlist; 
	var list = enumStruct(me);
	var found = false;
	for (j=0; j<list.length; ++j) {
		if (list[j].status == BEING_BUILT) {
			if (dangerLevel(list[j].x,list[j].y)<=0) {
				droidlist=sortByDistance(list[j].x,list[j].y,enumGroup(oilerGroup));
				var base=isBaseStructure(list[j]);
				for (var i=0; i<droidlist.length && (!found || base); ++i)	
					if (!isTruckBusy(droidlist[i]) && (droidlist[i].order != DORDER_MOVE))
						if (distBetweenTwoPoints(droidlist[i].x,droidlist[i].y,list[j].x,list[j].y)<BASE_SIZE)
							orderDroidObj(droidlist[i], DORDER_HELPBUILD, list[j]), found=true;
				if (base) {
					if (!enumStruct(me,generator).length)
						if (list[j].stattype == RESOURCE_EXTRACTOR || list[j].stattype == DEFENSE)
							break; //Don't assist in building derricks or defenses at the start of a game.
					droidlist=enumGroup(builderGroup);
					for (var i=0; i<droidlist.length; ++i)	
						if (!isTruckBusy(droidlist[i]))
							orderDroidObj(droidlist[i], DORDER_HELPBUILD, list[j]), found=true;
				}
			}
		}
		if (found) return true;
	}
	return found;
}

// build additional power generators when necessary
function buildAdditionalGenerators() {
	var genCount = enumStruct(me, generator).length;
	var derrickCount = enumStruct(me, derrick).length;
	if (derrickCount >= 4*genCount) {
		if (isStructureAvailable(generator,me)) {
			if (UNUSUAL_SITUATION==0) {
				if (playerPower(me) < LOW_POWER) {
					return buildBasicStructure(generator,0,1);
				}
			}
		}
	}
	return false;	
}

// build research modules
function researchRush() {
	if (!earlyGame(personality.PEACE_TIME))
		return false;
	var list = enumStruct(me, lab);
	for (var i=0; i<list.length; ++i) 
		if (list[i].modules < 1) 
			if (buildModule(list[i], rmod)) 
				return true;
	return false;
}

// build power modules
function upgradeGenerators() {
	var list = enumStruct(me, generator);
	for (var i=0; i<list.length; ++i) 
		if (list[i].modules < 1) 
			if (buildModule(list[i], pmod)) 
				return true;
	return false;
}

// build different sorts of modules
function upgradeStructures() {
	// don't build factory and research modules too much
	// if low on power
	if (outOfPower())
		return false;
	if (playerPower(me) < LOW_POWER)
		if (random(5)>0)
			return false;
	var list = enumStruct(me, factory);
	for (var i=0; i<list.length; ++i) 
		if (list[i].modules < 2) 
			if (buildModule(list[i], fmod)) 
				return true;
	list = enumStruct(me, lab);
	for (var i=0; i<list.length; ++i) 
		if (list[i].modules < 1) 
			if (buildModule(list[i], rmod)) 
				return true;
	list = enumStruct(me, vtolfac);
	for (var i=0; i<list.length; ++i) 
		if (list[i].modules < 2) 
			if (buildModule(list[i], fmod)) 
				return true;
	return false;
}

//Capture oil resources that are close to the base using builderGroup.
function secureBaseOil() {
	if (!enumStruct(me,generator).length)
		if (difficulty!=EASY)
			return false;
	var oils = enumFeature(-1,oilres);
	if (oils.length <= 0)
		return false; //No oils to secure.
	var minDist = Infinity;
	var dist = Infinity;
	var nearestOil;
	for (var i=0; i<oils.length; ++i) {
		dist=distBetweenTwoPoints(basePosition.x,basePosition.y,oils[i].x,oils[i].y);
		if (dist<minDist) {
			minDist=dist;
			nearestOil = oils[i];
		}
	}
	if (minDist>BASE_SIZE)
		return false;
	else {
		var droidlist = enumGroup(builderGroup);
		if (droidlist.length == 0) 
			return false;
		for (var i=0; i<droidlist.length; ++i) {
			if (!isTruckBusy(droidlist[i]))
				orderDroidBuild(droidlist[i], DORDER_BUILD, derrick, nearestOil.x, nearestOil.y);
		}
		return true;
	}
}

// a constructor for Derrick Stat object
function constructDerrickStat(pos) {
	this.x = pos.x; 
	this.y = pos.y;
	this.numBuilders = 0; // >0 if some trucks were sent recently to build derrick here
	this.hotness = -1; // the amount of times this derrick was rebuilt
	this.value = 0; //How many other oil resources are close to this resource.
	var list = enumFeature(-1,oilres);
	for (var i=0; i<11; ++i) {
		list=list.concat(enumStruct(i,derrick));
	}
	for (i=0; i<list.length; ++i) {
		for (var j=0; j<list.length; ++j) {
			if (distBetweenTwoPoints(list[i].x,list[i].y,list[j].x,list[j].y)<=BASE_SIZE/2)
				this.value+=1;
		}
	}
}

// finds an appropriate derrickStats entry for derrick
function findDerrickStat(obj) {
	if (typeof(derrickStats)=="undefined")
		derrickStats=[];
	for (var i=0; i<derrickStats.length; ++i) {
		if (derrickStats[i].x == obj.x && derrickStats[i].y == obj.y)
			return i;
	}
	var n=derrickStats.length;
	derrickStats[n]=new constructDerrickStat(obj);
	return n;
}

// update all group's targets periodically
// called from relaxStats
function updateGroupTargets() {
	if (typeof(updateGroupTargets.lastGroup)=="undefined")
		updateGroupTargets.lastGroup=0;
	if (groupInfo[updateGroupTargets.lastGroup].targetAge<=0) {
		getGroupTarget(updateGroupTargets.lastGroup);
	}
	++updateGroupTargets.lastGroup;
	if (updateGroupTargets.lastGroup==NUM_GROUPS)
		updateGroupTargets.lastGroup=0;
	else
		queue("updateGroupTargets",100); // update next group
}

// forgets that a derrick was already ordered
// forgets that the group doesn't currently want to regroup
// forgets that some factories are occupied with producing trucks
function relaxStats() {
	if (typeof(derrickStats)=="undefined")
		derrickStats=[];
	for (var i=0; i<derrickStats.length; ++i) {
		if (derrickStats[i].numBuilders>0)
			--derrickStats[i].numBuilders;
	}
	for (var i=0; i<NUM_GROUPS; ++i) {
		if (groupInfo[i].delayregroup>0) 
			--groupInfo[i].delayregroup;
		if (groupInfo[i].delaystaticregroup>0) 
			--groupInfo[i].delaystaticregroup;
		if (groupInfo[i].targetAge>0) 
			--groupInfo[i].targetAge;
	}
	updateGroupTargets();
}

function findHottestDerrick() {
	var oils=enumStruct(me,derrick);
	for (var i=0; i<maxPlayers; ++i)
		if (allianceExistsBetween(me,i))
			oils=oils.concat(enumStruct(i,derrick)); //List allied derricks as well.
	var list=[];
	for (var i=0; i<oils.length; ++i) {
		var h=(derrickStats[findDerrickStat(oils[i])].hotness*2)+1; //List "cold" derricks as well.
		for (var j=0; j<h; ++j) {
			list[list.length]=oils[i];
		}
	}
	if (list.length>0) 
		return list[random(list.length)];
	else
		return oils[random(oils.length)];
}

function findMostValuableDerrick() {
	var oils=enumStruct(me,derrick);
	for (var i=0; i<maxPlayers; ++i)
		if (allianceExistsBetween(me,i))
			oils=oils.concat(enumStruct(i,derrick)); //List allied derricks as well.
	var list=[];
	for (var i=0; i<oils.length; ++i) {
		var h=(derrickStats[findDerrickStat(oils[i])].value);
		for (var j=0; j<h; ++j) {
			list[list.length]=oils[i];
		}
	}
	if (list.length>0) 
		return list[random(list.length)];
	else
		return oils[random(oils.length)];
}

// tell idle units to pick up oil drums
function pickupOilDrums() {
	if (typeof(pickupOilDrums.lastCall)!="undefined" && gameTime - pickupOilDrums.lastCall < 20000) // don't call this too often
		return;
	pickupOilDrums.lastCall = gameTime;
	if (typeof(DORDER_RECOVER)=="undefined") // HACK: this constant isn't exposed for scripts yet
		DORDER_RECOVER=33;
	if (status == STATUS_PANIC)
		return;
	var drums=enumFeature(-1,oildrum);
	drums=drums.concat(enumFeature(-1,artifact));
	var list=enumGroup(oilerGroup);
	list=list.concat(enumGroup(defendGroup));
	list=list.concat(enumGroup(harassGroup));
	for (var i=0; i<list.length; ++i) {
		var droid=list[i];
		if (isTruckBusy(droid))
			continue;
		var minDist=Infinity, minObj;
		for (var j=0; j<drums.length; ++j) {
			var drum=drums[j];
			if (droidCanReach(droid,drum.x,drum.y)) {
				var dist=distBetweenTwoPoints(drum.x,drum.y,droid.x,droid.y);
				if (dist < minDist) {
					minDist=dist;
					minObj=drum;
				}
			}
		}
		if (typeof(minObj)!="undefined")
			if (dangerLevel(minObj.x,minObj.y)<=0) {
				orderDroidObj(droid,DORDER_RECOVER,minObj);
				return;
			}
	}
}

// order free oil-oriented trucks to build oil derricks
function huntForOil() {
	var droidlist = enumGroup(oilerGroup);
	var list = enumFeature(-1, oilres);
	var oillist = [];
	if (typeof(huntForOil.cachedOilList)=="undefined" || gameTime - huntForOil.cacheUpdateTime > 10000) {
		for (var j=0; j<list.length; ++j) {
			var stat=findDerrickStat(list[j]);
			if (derrickStats[stat].numBuilders<=0) {
				var danger=false;
				if (list.length > 20) {
					if (!safeDest(me,list[j].x,list[j].y))
						danger=true;
				} else {
					if (dangerLevel(list[j].x,list[j].y)>0)
						danger=true;
					else {
						if (isNearHotspot(list[j].x,list[j].y)) {
							danger=true;
							break;
						}
					}
				}
				if (!danger || random(15)==0) {
					var tmp=getNearestCCPlayer(list[j].x,list[j].y,0); // friendly AIs shouldn't capture your oil
					if (typeof(tmp)=="undefined" || tmp==me) {
						var n=oillist.length;
						oillist[n]=list[j];
						oillist[n].stat=stat;
					}
				}
			}
		}
		huntForOil.cachedOilList = oillist;
		huntForOil.cacheUpdateTime = gameTime;
	} else {
		oillist = huntForOil.cachedOilList;
	}
	if (oillist.length <= 0) { // not much to do here ...
		pickupOilDrums();
		return;
	}
	for (var i=0; i<droidlist.length; ++i) if (!isTruckBusy(droidlist[i])) {
		var range1=Infinity;
		var k=-1;
		for (var j=0; j<oillist.length; ++j) {
			if (derrickStats[oillist[j].stat].numBuilders<=0) {
				if (droidCanReach(droidlist[i],oillist[j].x,oillist[j].y)) {
					var range2=distBetweenTwoPoints(droidlist[i].x, droidlist[i].y, oillist[j].x, oillist[j].y);
					if (range2<range1) {
						range1=range2;
						k=j;
					}
				}
			}
		}
		if (k!=-1) {
			orderDroidBuild(droidlist[i], DORDER_BUILD, derrick, oillist[k].x, oillist[k].y);
			//make sure not too many trucks build the same oil
			// 5 here makes sure that at least five calls of relaxStats() occur before the stat is relaxed
			// which takes 40 to 50 seconds
			derrickStats[findDerrickStat(oillist[k])].numBuilders = 5; 
		} //else {
			//return to base if no oils are available for capturing
			/*if (droidlist[i].order != DORDER_MOVE)
				returnToBase(droidlist[i]);*/
		//}
	}

}


/* Researching technologies */


// make a particular lab follow a certain research path
// even though pursueResearch() accepts paths, 
// its behaviour seems to be less deterministic than this
function followResearchPath(lab,path) {
	for (var j=0; j<path.length; ++j) {
		//We should skip over technologies that appear on any of our blacklists.
		var success=true;
		if (RATE_TANK==0) //Hover researching blacklist.
			for (var i=0; i<personality.hoverBlacklist.length; ++i)
				if (path[j]==personality.hoverBlacklist[i])
					success=false;
		if (success)
			if (pursueResearch(lab,path[j]))
				return true;
		}
	return false;
}

// come up with something
function doResearch() {
	if (!dontRunOutOfPower())
		return;
	var lablist=enumStruct(me, RESEARCH_LAB);
	for (var i=0; i<lablist.length; ++i) if (structureReady(lablist[i])) {
		if (researchBlocked()) 
			return;
		if (outOfPower() && i>2) //Limit research to 3 labs if already out of power.
			return;
		if (followResearchPath(lablist[i],forcedResearchPath))
			continue;
		if (RATE_TANK==0 && !iHaveHovers()) { // we seriously need hovers as soon as possible
			if (followResearchPath(lablist[i],personality.researchPathHover)) {
				continue;
			}
		}
		if (MIN_DEFENDERS_AA>0) { //We really need anti-air units as soon as we see enemy VTOLs.
			if (followResearchPath(lablist[i],personality.researchPathFundamentalAA))
				continue;
		}
		if (!followResearchPath(lablist[i],personality.researchPathPrimary)) {
			// pick some branchy research if the primary research path is finished
			if (playerPower(me) < LOW_POWER) { //Focus on researching power upgrades if low on power.
				if (followResearchPath(lablist[i],personality.researchPathEconomy))
					continue;
			}
			//If an exotic technology (that requires research into more than 1 research line) is available, then pursue it.
			var reslist = enumResearch();
			if (reslist.length>0) {
				var success = false;
				for (var j=0; j<reslist.length; ++j) {
					for (var k=0; k<personality.researchPathExotic.length; ++k) {
						if (reslist[j].name == personality.researchPathExotic[k]) {
							success = pursueResearch(lablist[i],reslist[j].name);
							break;
						}
					if (success)
						break;
					}
				}
				if (success)
					continue;
			}
			var economy=enumStruct(me,derrick).length;
			if (random(140+economy*4)>100) // with 40 oils, it will pick fundamental research in around 2/3 labs
				if (followResearchPath(lablist[i],personality.researchPathFundamental)) {
					continue;
				}
			var ret=enemyCountBalance(-1);
			// vtols can be really scary, so give a bit more weight to AA research
			//ret.vtol = ret.vtol * 3 + ((earlyGame(20) && !HIGH_TECH_START==1)?0:1);
			// don't forget completely about borgs
			//ret.borg += 5;
			if (difficulty == EASY || difficulty == MEDIUM) {
				ret.tank = 4;
				ret.borg = 4;
				ret.vtol = 1;
				ret.defs = 1;
			}
			var total=ret.tank+ret.borg+ret.vtol+ret.defs;
			if (random(total)<ret.borg)
				if (followResearchPath(lablist[i],personality.researchPathAP)) {
					continue;
				}
			total=ret.tank+ret.vtol+ret.defs;
			if (random(total)<ret.tank)
				if (followResearchPath(lablist[i],personality.researchPathAT)) {
					continue;
				}
			total=ret.vtol+ret.defs;
			if (random(total)<ret.vtol)
				if (followResearchPath(lablist[i],personality.researchPathAA)) {
					continue;
				}
			if (followResearchPath(lablist[i],personality.researchPathAB))
				continue;
			//If there are still some free slots available, then pursue fundamental research.
			if (followResearchPath(lablist[i],personality.researchPathFundamental))
				continue;
			if (followResearchPath(lablist[i],personality.researchPathEconomy))
				continue;
			//If adaptation has failed, then stop considering it.
			var r = random(4);
			var success = false;
			for (var j=0; j<4; ++j) {
				switch (r) {
				case 0:
					success = followResearchPath(lablist[i],personality.researchPathAP);
					break;
				case 1:
					success = followResearchPath(lablist[i],personality.researchPathAT);
					break;
				case 2:
					success = followResearchPath(lablist[i],personality.researchPathAB);
					break;
				case 3:
					success = followResearchPath(lablist[i],personality.researchPathAA);
					break;
				}
				if (success)
					break;
				r+=1;
				if (r>=4)
					r-=4;
			}
			if (success)
				continue;
			//Start pursuing exotic technology if no other techs are available.
			if (followResearchPath(lablist[i],personality.researchPathExotic))
				continue;
			//If all else fails, then just research whatever hasn't been researched yet, but ONLY if there's a surplus of power.
			//Note however that an EASY AI will keep on trying to research stuff even if low on power.
			var reslist = enumResearch();
			if (reslist.length > 0 && (playerPower(me) > LOW_POWER || difficulty==EASY))
			{
				var idx = random(reslist.length);
				pursueResearch(lablist[i], reslist[idx].name);
			}
		//In order for the function to successfully limit which research labs research stuff based on
		//available power, we'll need to queue up and terminate the function here and now,
		//and let structureReady() find the next available research lab.
		//queue("doResearch",100);
		//return;
		} 
	}
}


/* Producing droids */	

// not sure this check is necessary ... why not
function structureReady(struct) {
	return structureIdle(struct) && (struct.status == BUILT);
}

function outOfPower() {
	/*if (typeof(queuedPower())!="undefined")
		return (playerPower(me)<=0 || queuedPower(me)>0);
	else*/
		return (playerPower(me)<=0);
}


// finds out which battle group contains this tank
function groupOfTank(tank) {
	for (i=0; i<NUM_GROUPS; ++i)
		if (tank.group == battleGroup[i])
			return i;
}

// pick a suitable group for a new-born truck
function addTruckToSomeGroup(truck) {
	if (groupSize(builderGroup) < personality.MIN_BUILDERS)
		groupAddDroid(builderGroup,truck);
	else if (groupSize(oilerGroup) < personality.MIN_OILERS && !earlyGame(personality.PEACE_TIME))
		groupAddDroid(oilerGroup,truck);
	else if (groupSize(builderGroup) < personality.MAX_BUILDERS)
		groupAddDroid(builderGroup,truck);
	else if (groupSize(oilerGroup) < personality.MAX_OILERS)
		groupAddDroid(oilerGroup,truck);
	else if (groupSize(builderGroup)+groupSize(oilerGroup)>ABS_MAX_TRUCKS)
		groupAddDroid(oilerGroup,truck)
	else if (random(2)==0)
		groupAddDroid(builderGroup,truck);
	else
		groupAddDroid(oilerGroup,truck);
	if (truck.group==oilerGroup && (truck.name=="Heavy Truck" || truck.name=="Medium Truck"))
		replaceTrucks(truck);
}

// checks if adding a tank of type defined by `at' parameter would
// bring some group closer to its AT/AP weapon balance
// at=true for AT tanks, at=false for AP tanks
function groupNeedsThisSortOfTank(gr,name) {
	var pt=groupCountBalance(gr);
	var tb=enemyCountBalance(groupInfo[gr].enemy);
	var ret=adaptationMatrix(pt.ap, pt.at, pt.aa, pt.ab, tb.tank, tb.borg, tb.vtol, tb.defs);
	if (name.indexOf("AP")>0)
		return ret.ap>0;
	if (name.indexOf("AT")>0)
		return ret.at>0;
	if (name.indexOf("AA")>0)
		return ret.aa>0;
	if (name.indexOf("AB")>0)
		return ret.ab>0;
}

// checks if adding a vtol of type defined by `at' parameter would
// bring some group closer to its AT/AP weapon balance
// at=true for AT tanks, at=false for AP tanks
function groupNeedsThisSortOfVtol(gr,name) {
	var pt=groupCountBalanceVtol(gr);
	var ret=adaptationMatrix(pt.ap, pt.at, pt.aa, pt.ab, RATE_AT, RATE_AP, RATE_AA, RATE_AB);
	if (name.indexOf("AP")>0)
		return ret.ap>0;
	if (name.indexOf("AT")>0)
		return ret.at>0;
	if (name.indexOf("AA")>0)
		return ret.aa>0;
	if (name.indexOf("AB")>0)
		return ret.ab>0;
}

// how many non-empty combat groups do we have?
function listNonEmptyGroups() {
	var ret=[];
	for (var i=0; i<NUM_GROUPS; ++i)
		if (groupSize(battleGroup[i])>0)
			ret[ret.length]=i;
	return ret;
}

// pick a suitable group for a new-born tank
function addTankToSomeGroup(tank) {
	forceReturnToBase(tank);
	if (tank.player!=me) {
		debug("BUG: Player ",me,"trying to put tank",tank.id,"in a group, but the tank belongs to player",tank.player);
		return;
	}
	if (((random(40)+(groupSize(harassGroup)*3)+10)<enumDroid(me).length || earlyGame(personality.PEACE_TIME)) && groupSize(harassGroup)<NUM_HARASS && tank.name.indexOf("AA")<=0) {
		groupAddDroid(harassGroup, tank);
		return;
	}
	if (tank.name.indexOf("AA")>0 && (groupSize(aaVtolGroup)+groupSize(aaGroup))<MIN_DEFENDERS_AA) {
		groupAddDroid(aaGroup, tank);
		return;
	}
	if (random(100)<countEnemyTeams()*3 && groupSize(defendGroup)<MAX_WARRIORS) {
		groupAddDroid(defendGroup, tank);
		return;
	}
	for (var i=0; i<NUM_GROUPS; ++i)  {
		if (typeof(groupInfo[i].enemy)=="undefined" || !isAnEnemy(groupInfo[i].enemy))
			groupInfo[i].enemy=findEnemy();
	}
	var maxCount=-1, maxIdx=-1;
	// find the biggest non-full group that needs this sort of tank
	for (var i=0; i<NUM_GROUPS; ++i)  {
		if (typeof(groupInfo[i].enemy)=="undefined")
			continue;
		if (!droidCanReach(tank,startPositions[groupInfo[i].enemy].x,startPositions[groupInfo[i].enemy].y))
			continue;
		var count=groupSize(battleGroup[i]);
		if (count<NUM_WARRIORS && count>0 && groupNeedsThisSortOfTank(i,tank.name))
			if (count > maxCount) {
				maxCount = count;
				maxIdx = i;
			}
	}
	if (maxIdx!=-1) {
		groupAddDroid(battleGroup[maxIdx], tank);
		return;
	}
	maxCount=-1, maxIdx=-1;
	// if nobody needs this sort of tank, find the biggest non-full group
	for (var i=0; i<NUM_GROUPS; ++i) {
		if (typeof(groupInfo[i].enemy)=="undefined")
			continue;
		if (!droidCanReach(tank,startPositions[groupInfo[i].enemy].x,startPositions[groupInfo[i].enemy].y))
			continue;
		var count=groupSize(battleGroup[i]);
		if (count<NUM_WARRIORS && count>0)
			if (count > maxCount) {
				maxCount = count;
				maxIdx = i;
			}
	}
	if (maxIdx!=-1) {
		groupAddDroid(battleGroup[maxIdx], tank);
		return;
	}
	// if all groups are full, fill until absolute maximum
	// but still consider AT/AP balance
	for (var i=0; i<NUM_GROUPS; ++i) {
		if (typeof(groupInfo[i].enemy)=="undefined")
			continue;
		if (!droidCanReach(tank,startPositions[groupInfo[i].enemy].x,startPositions[groupInfo[i].enemy].y))
			continue;
		var count=groupSize(battleGroup[i]);
		if (count<MAX_WARRIORS && count>0 && groupNeedsThisSortOfTank(i,tank.name)) {
			groupAddDroid(battleGroup[i], tank);
			return;
		}
	}
	// if considering AT/AP balance failed, stop considering it
	for (var i=0; i<NUM_GROUPS; ++i) {
		if (typeof(groupInfo[i].enemy)=="undefined")
			continue;
		if (!droidCanReach(tank,startPositions[groupInfo[i].enemy].x,startPositions[groupInfo[i].enemy].y))
			continue;
		var count=groupSize(battleGroup[i]);
		if (count<MAX_WARRIORS && count>0) {
			groupAddDroid(battleGroup[i], tank);
			return;
		}
	}
	var ret=listNonEmptyGroups();
	if (ret.length < countEnemies()) {
		// if we just have no groups yet, put the tank into a new group
		for (var i=0; i<NUM_GROUPS; ++i) 
			if (groupSize(battleGroup[i])==0) {
				groupAddDroid(battleGroup[i], tank);
				return;
			}
	} else {
		// huh, still no success? come on, just put it anywhere already
		groupAddDroid(battleGroup[ret[random(ret.length)]], tank);
	}
}

// pick a suitable group for a new-born vtol
function addVtolToSomeGroup(vtol) {
	if (vtol.name.indexOf("AA")>0) {
		groupAddDroid(aaVtolGroup, vtol);
		return;
	}
	if (vtol.name.indexOf("AB")>0) {
		groupAddDroid(abVtolGroup, vtol);
		return;
	}
	// find the biggest non-full group
	var maxCount=-1, maxIdx=-1;
	for (var i=0; i<NUM_VTOL_GROUPS; ++i) {
		var count=groupSize(vtolGroup[i])
		if (count<NUM_VTOLS)
			if (count > maxCount) {
				maxCount = count;
				maxIdx = i;
			}
	}
	if (maxIdx!=-1) {
		groupAddDroid(vtolGroup[maxIdx], vtol);
		return;
	}
	// if all groups are full, fill until maximum
	for (var i=0; i<NUM_VTOL_GROUPS; ++i)
		if (groupSize(vtolGroup[i])<MAX_VTOLS) {
			groupAddDroid(vtolGroup[i], vtol);
			return;
		}
	// if all groups are at maximum, add a new group
	if (NUM_VTOL_GROUPS < MAX_VTOL_GROUPS) {
		vtolGroup[NUM_VTOL_GROUPS]=newGroup();
		groupAddDroid(vtolGroup[NUM_VTOL_GROUPS], vtol);
		++NUM_VTOL_GROUPS;
		return;
	}
	// if too many groups, just put it somewhere
	groupAddDroid(vtolGroup[random(NUM_VTOL_GROUPS)], vtol);
}

// checks if we actually need more trucks
function needMoreTrucks() {
	if (status==STATUS_PANIC && personality.THIS_AI_MAKES_TANKS)
		return false;
	if (iHaveHovers()) {
		if (groupSize(oilerGroup)+groupSize(builderGroup)>=ABS_MAX_HOVER_TRUCKS)
			return false;
	} else {
		if (groupSize(oilerGroup)+groupSize(builderGroup)>=ABS_MAX_TRUCKS)
			return false;
	}
	var weHaveHoverTrucks=false;
	var list=enumGroup(oilerGroup);
	for (var i=0; i<list.length; ++i) {
		if (list[i].name.indexOf("Hover")>-1) {
			weHaveHoverTrucks = true;
			break;
		}
	}
	if (iHaveHovers() && !weHaveHoverTrucks)
		return true;
	if (groupSize(oilerGroup) >= personality.MAX_OILERS && groupSize(builderGroup) >= personality.MAX_BUILDERS)
		if (playerPower(me)<LOW_POWER || enumDroid(me,DROID_WEAPON).length < personality.MIN_DEFENDERS)
			return false;
	if (groupSize(oilerGroup) >= personality.MIN_OILERS && groupSize(builderGroup) >= personality.MIN_BUILDERS)
		if (enumDroid(me,DROID_WEAPON).length < personality.MIN_DEFENDERS)
			return false;
	return true;
}

//Send a droid to the nearest factory or repair facility for recycling.
function recycleDroid(droid) {
	if (typeof(DORDER_RECYCLE) == "undefined") //Note: DORDER_RECYCLE is only present in 3.2 builds.
		return false;
	var scrapyards=enumStruct(me,FACTORY).concat(enumStruct(me,CYBORG_FACTORY),enumStruct(me,VTOL_FACTORY),enumStruct(me,REPAIR_FACILITY));
	if (scrapyards.length==0)
		return false; //No valid structures detected.
	var struct=findNearestObject(droid.x,droid.y,scrapyards);
	return orderDroidObj(droid,DORDER_RECYCLE,struct);
}

//Sell old trucks to make room for larger, more advanced ones.
function recycleTrucks() {
	var trucks=enumDroid(me,DROID_CONSTRUCT).concat(enumDroid(me,DROID_CYBORG_CONSTRUCT));
	if (trucks.length>=ABS_MAX_TRUCKS) { //Sell Combat Engineers 1st.
		for (var i=0; i<trucks.length; ++i)
			if (trucks[i].droidType==DROID_CYBORG_CONSTRUCT)
				if (recycleDroid(trucks[i]))
					return true;
		for (var i=0; i<trucks.length; ++i) //Sell Light Trucks next to make room for Medium Trucks and Hover Trucks.
			if (trucks[i].name!="Medium Truck" && trucks[i].name!="Heavy Truck" && trucks[i].name!="Light Hover Truck")
				if (recycleDroid(trucks[i]))
					return true;
		for (var i=0; i<trucks.length; ++i) //Sell Medium Trucks to make room for Heavy Trucks and Hover Trucks.
			if (trucks[i].name!="Heavy Truck" && trucks[i].name!="Light Hover Truck")
				if (recycleDroid(trucks[i]))
					return true;
		//If we can't recycle our trucks, then we can instead try to reassign our least valuable builderGroup trucks to oilerGroup,
		//and reassign our most valuable oilerGroup trucks to builderGroup.
		if (replaceTrucks())
			return true;
	}
	return false;
}

//Swap out old, smaller builderGroup trucks with new, larger oilerGroup trucks.
function replaceTrucks(newBuilder) {
	trucks=enumGroup(builderGroup);
	if (trucks.length>0) {
		for (var i=0; i<trucks.length; ++i)
			if (trucks[i].droidType==DROID_CYBORG_CONSTRUCT)
				var newOiler=trucks[i];
		if (typeof(newOiler) == "undefined")
			for (var i=0; i<trucks.length; ++i)
				if (trucks[i].name!="Medium Truck" && trucks[i].name!="Heavy Truck")
					var newOiler=trucks[i];
		if (typeof(newOiler) == "undefined")
			for (var i=0; i<trucks.length; ++i)
				if (trucks[i].name!="Heavy Truck")
					var newOiler=trucks[i];
		if (typeof(newOiler) != "undefined") {
			//Now that we've found a builderGroup truck to move to oilerGroup,
			//we need to find an oilerGroup truck to move to builderGroup if one wasn't already specified.
			if (typeof(newBuilder) == "undefined") {
				trucks=enumGroup(oilerGroup);
				if (trucks.length>0) {
					for (var i=0; i<trucks.length; ++i)
						if (trucks[i].name=="Heavy Truck")
							var newBuilder=trucks[i];
					if (typeof(newBuilder) == "undefined")
						for (var i=0; i<trucks.length; ++i)
							if (trucks[i].name=="Medium Truck" && newOiler.name!="Medium Truck")
								var newBuilder=trucks[i];
				}
			}
			//Now that we've found a suitable oilerGroup truck to move to builderGroup, let's proceed with swapping their roles.
			if (typeof(newBuilder) != "undefined") {
				if (newBuilder.name != newOiler.name) {
					groupAddDroid(builderGroup,newBuilder);
					groupAddDroid(oilerGroup,newOiler);
					return true;
				}
			}
		}
	}
	return false;
}

// produce a tank with an anti-personnel weapon in some factory
function produceAPTank(struct) {
	if (!BUILD_TANKS)
		return false;
	var name = "AP Tank";
	var prop = personality.tankPropulsions;
	if (typeof(personality.tankPropulsionsAP)!="undefined")
		prop=personality.tankPropulsionsAP;
	var body = personality.tankBodies;
	if (typeof(personality.tankBodiesAP)!="undefined")
		body=personality.tankBodiesAP;
	if (random(RATE_TANK+RATE_HOVER) < RATE_HOVER)
		prop=standardTankPropulsionsHover;
	var weapon1 = personality.apWeapons;
	var weapon2 = weapon1;
	var weapon3 = weapon2;
	var iap1=random(weapon1.length);
	var iap2=random(weapon2.length);
	var iap3=random(weapon3.length);
	if (difficulty != EASY) {
		if (random(RATE_AP_LIGHT+RATE_AP_MEDIUM+RATE_AP_HEAVY+RATE_AP_SUPER) < RATE_AP_SUPER && struct.modules>=2) {
			if (random(RATE_AP_SUPER_A+RATE_AP_SUPER_B) < RATE_AP_SUPER_A) {
				weapon1 = personality.apWeaponsSuper;
				weapon2 = personality.apWeaponsLight;
				iap1=random(weapon1.length);
				iap2=random(weapon2.length);
				body = personality.tankBodiesSuper;
				name = "Super AP Tank";
			}
			else {
				weapon1 = personality.apWeaponsHeavy;
				weapon2 = weapon1;
				iap1=random(weapon1.length);
				iap2=random(weapon2.length);
				body = personality.tankBodiesSuper;
				name = "Super AP Tank";
			}
		}
		else if (random(RATE_AP_LIGHT+RATE_AP_MEDIUM+RATE_AP_HEAVY) < RATE_AP_HEAVY && struct.modules>=2) {
			weapon1 = personality.apWeaponsHeavy;
			iap1=random(weapon1.length);
			body = personality.tankBodiesHeavy;
			name = "Heavy AP Tank";
		}
		else if (random(RATE_AP_LIGHT+RATE_AP_MEDIUM) < RATE_AP_MEDIUM && struct.modules>=1) {
			weapon1 = personality.apWeaponsMedium;
			iap1=random(weapon1.length);
			body = personality.tankBodiesMedium;
			name = "Medium AP Tank";
		}
		else {
			weapon1 = personality.apWeaponsLight;
			iap1=random(weapon1.length);
			body = personality.tankBodiesLight;
			name = "Light AP Tank";
		}
	}
	var ip=random(prop.length);
	var ib=random(body.length);
	return buildDroid(struct, name, body[ib], prop[ip], "", DROID_WEAPON, weapon1[iap1], weapon2[iap2], weapon3[iap3]);
}

// produce a tank with an anti-tank weapon in some factory
function produceATTank(struct) {
	if (!BUILD_TANKS)
		return false;
	var name = "AT Tank";
	var prop = personality.tankPropulsions;
	if (typeof(personality.tankPropulsionsAT)!="undefined")
		prop=personality.tankPropulsionsAT;
	var body = personality.tankBodies;
	if (typeof(personality.tankBodiesAT)!="undefined")
		body=personality.tankBodiesAT;
	if (random(RATE_TANK+RATE_HOVER) < RATE_HOVER)
		prop=standardTankPropulsionsHover;
	var weapon1 = personality.atWeapons;
	var weapon2 = weapon1;
	var weapon3 = weapon2;
	var iat1=random(weapon1.length);
	var iat2=random(weapon2.length);
	var iat3=random(weapon3.length);
	if (difficulty != EASY) {
		if (random(RATE_AT_LIGHT+RATE_AT_MEDIUM+RATE_AT_HEAVY+RATE_AT_SUPER) < RATE_AT_SUPER && struct.modules>=2) {
			if (random(RATE_AT_SUPER_A+RATE_AT_SUPER_B) < RATE_AT_SUPER_A) {
				weapon1 = personality.atWeaponsSuper;
				weapon2 = personality.atWeaponsLight;
				iat1=random(weapon1.length);
				iat2=random(weapon2.length);
				body = personality.tankBodiesSuper;
				name = "Super AT Tank";
			}
			else {
				weapon1 = personality.atWeaponsHeavy;
				weapon2 = weapon1;
				iat1=random(weapon1.length);
				iat2=random(weapon2.length);
				body = personality.tankBodiesSuper;
				name = "Super AT Tank";
			}
		}
		else if (random(RATE_AT_LIGHT+RATE_AT_MEDIUM+RATE_AT_HEAVY) < RATE_AT_HEAVY && struct.modules>=2) {
			weapon1 = personality.atWeaponsHeavy;
			iat1=random(weapon1.length);
			body = personality.tankBodiesHeavy;
			name = "Heavy AT Tank";
		}
		else if (random(RATE_AT_LIGHT+RATE_AT_MEDIUM) < RATE_AT_MEDIUM && struct.modules>=1) {
			weapon1 = personality.atWeaponsMedium;
			iat1=random(weapon1.length);
			body = personality.tankBodiesMedium;
			name = "Medium AT Tank";
		}
		else {
			weapon1 = personality.atWeaponsLight;
			iat1=random(weapon1.length);
			body = personality.tankBodiesLight;
			name = "Light AT Tank";
		}
	}
	var ip=random(prop.length);
	var ib=random(body.length);
	return buildDroid(struct, name, body[ib], prop[ip], "", DROID_WEAPON, weapon1[iat1], weapon2[iat2], weapon3[iat3]);
}

// produce a tank with an anti-building weapon in some factory
function produceABTank(struct) {
	if (!BUILD_TANKS)
		return false;
	var name = "AB Tank";
	var prop = personality.tankPropulsions;
	if (typeof(personality.tankPropulsionsAB)!="undefined")
		prop=personality.tankPropulsionsAB;
	var body = personality.tankBodies;
	if (typeof(personality.tankBodiesAB)!="undefined")
		body=personality.tankBodiesAB;
	if (random(RATE_TANK+RATE_HOVER) < RATE_HOVER)
		prop=standardTankPropulsionsHover;
	var weapon1 = personality.abWeapons;
	var weapon2 = weapon1;
	var weapon3 = weapon2;
	var iab1=random(weapon1.length);
	var iab2=random(weapon2.length);
	var iab3=random(weapon3.length);
	if (difficulty != EASY) {
		if (random(RATE_AB_LIGHT+RATE_AB_MEDIUM+RATE_AB_HEAVY+RATE_AB_SUPER) < RATE_AB_SUPER && struct.modules>=2) {
			if (random(RATE_AB_SUPER_A+RATE_AB_SUPER_B) < RATE_AB_SUPER_A) {
				weapon1 = personality.abWeaponsSuper;
				weapon2 = personality.abWeaponsLight;
				iab1=random(weapon1.length);
				iab2=random(weapon2.length);
				body = personality.tankBodiesSuper;
				name = "Super AB Tank";
			}
			else {
				weapon1 = personality.abWeaponsHeavy;
				weapon2 = weapon1;
				iab1=random(weapon1.length);
				iab2=random(weapon2.length);
				body = personality.tankBodiesSuper;
				name = "Super AB Tank";
			}
		}
		else if (random(RATE_AB_LIGHT+RATE_AB_MEDIUM+RATE_AB_HEAVY) < RATE_AB_HEAVY && struct.modules>=2) {
			weapon1 = personality.abWeaponsHeavy;
			iab1=random(weapon1.length);
			body = personality.tankBodiesHeavy;
			name = "Heavy AB Tank";
		}
		else if (random(RATE_AB_LIGHT+RATE_AB_MEDIUM) < RATE_AB_MEDIUM && struct.modules>=1) {
			weapon1 = personality.abWeaponsMedium;
			iab1=random(weapon1.length);
			body = personality.tankBodiesMedium;
			name = "Medium AB Tank";
		}
		else {
			weapon1 = personality.abWeaponsLight;
			iab1=random(weapon1.length);
			body = personality.tankBodiesLight;
			name = "Light AB Tank";
		}
	}
	var ip=random(prop.length);
	var ib=random(body.length);
	return buildDroid(struct, name, body[ib], prop[ip], "", DROID_WEAPON, weapon1[iab1], weapon2[iab2], weapon3[iab3]);
}

// produce a tank with an anti-air weapon in some factory
function produceAATank(struct) {
	if (!BUILD_TANKS)
		return false;
	var name = "AA Tank";
	var prop = personality.tankPropulsions;
	if (typeof(personality.tankPropulsionsAA)!="undefined")
		prop=personality.tankPropulsionsAA;
	var body = personality.tankBodies;
	if (typeof(personality.tankBodiesAA)!="undefined")
		body=personality.tankBodiesAA;
	if (random(RATE_TANK+RATE_HOVER) < RATE_HOVER)
		prop=standardTankPropulsionsHover;
	var weapon1 = personality.aaWeapons;
	var weapon2 = weapon1;
	var weapon3 = weapon2;
	var iaa1=random(weapon1.length);
	var iaa2=random(weapon2.length);
	var iaa3=random(weapon3.length);
	if (difficulty != EASY) {
		if (random(RATE_AA_LIGHT+RATE_AA_MEDIUM+RATE_AA_HEAVY+RATE_AA_SUPER) < RATE_AA_SUPER && struct.modules>=2) {
			if (random(RATE_AA_SUPER_A+RATE_AA_SUPER_B) < RATE_AA_SUPER_A) {
				weapon1 = personality.aaWeaponsSuper;
				weapon2 = personality.aaWeaponsLight;
				iaa1=random(weapon1.length);
				iaa2=random(weapon2.length);
				body = personality.tankBodiesSuper;
				name = "Super AA Tank";
			}
			else {
				weapon1 = personality.aaWeaponsHeavy;
				weapon2 = weapon1;
				iaa1=random(weapon1.length);
				iaa2=random(weapon2.length);
				body = personality.tankBodiesSuper;
				name = "Super AA Tank";
			}
		}
		else if (random(RATE_AA_LIGHT+RATE_AA_MEDIUM+RATE_AA_HEAVY) < RATE_AA_HEAVY && struct.modules>=2) {
			weapon1 = personality.aaWeaponsHeavy;
			iaa1=random(weapon1.length);
			body = personality.tankBodiesHeavy;
			name = "Heavy AA Tank";
		}
		else if (random(RATE_AA_LIGHT+RATE_AA_MEDIUM) < RATE_AA_MEDIUM && struct.modules>=1) {
			weapon1 = personality.aaWeaponsMedium;
			iaa1=random(weapon1.length);
			body = personality.tankBodiesMedium;
			name = "Medium AA Tank";
		}
		else {
			weapon1 = personality.aaWeaponsLight;
			iaa1=random(weapon1.length);
			body = personality.tankBodiesLight;
			name = "Light AA Tank";
		}
	}
	var ip=random(prop.length);
	var ib=random(body.length);
	return buildDroid(struct, name, body[ib], prop[ip], "", DROID_WEAPON, weapon1[iaa1], weapon2[iaa2], weapon3[iaa3]);
}

// produce a tank designed specifically for harassing opponents
function produceHarasser(struct) {
	if (!BUILD_TANKS || groupSize(harassGroup)>=NUM_HARASS)
		return false;
	var name = "Harasser Tank";
	var prop = personality.harassPropulsions;
	var body = personality.harassBodies;
	if (random(RATE_TANK+RATE_HOVER) < RATE_HOVER)
		prop=standardTankPropulsionsHover;
	var weapon1 = personality.harassWeapons;
	var weapon2 = weapon1;
	var weapon3 = weapon2;
	var iha1=random(weapon1.length);
	var iha2=random(weapon2.length);
	var iha3=random(weapon3.length);
	var ip=random(prop.length);
	var ib=random(body.length);
	return buildDroid(struct, name, body[ib], prop[ip], "", DROID_WEAPON, weapon1[iha1], weapon2[iha2], weapon3[iha3]);
}

//Produce a tank equipped with a sensor.
function produceSensorTank(struct) {
	var sensortanks=0, cbtanks=0, detectortanks=0;
	for (var gr=0; gr<NUM_GROUPS; ++gr) {
		var ret=groupCountBalance(gr);
		sensortanks += ret.se;
		cbtanks += ret.cb;
		detectortanks += ret.de;
	}
	var r=random(3);
	var success=false;
	for (var i=0; i<3; ++i) {
		switch (r) {
		case 0:
			if (!personality.THIS_AI_MAKES_TANKS || cbtanks < MAX_CB_TANKS) {
				var prop=sensorPropulsions;
				if (RATE_TANK == 0)
					prop=hoverPropulsions;
				success = buildDroid(struct, "CB Tank", sensorBodies, prop, "", DROID_SENSOR, cbTurrets);
			}
			break;
		case 1:
			if (!personality.THIS_AI_MAKES_TANKS || sensortanks < MAX_SENSOR_TANKS) {
				var prop=sensorPropulsions;
				if (RATE_TANK == 0)
					prop=hoverPropulsions;
				success = buildDroid(struct, "Sensor Tank", sensorBodies, prop, "", DROID_SENSOR, sensorTurrets);
			}
			break;
		case 2:
			if (!personality.THIS_AI_MAKES_TANKS || detectortanks < MAX_DETECTOR_TANKS) {
				var prop=sensorPropulsions;
				if (RATE_TANK == 0)
					prop=hoverPropulsions;
				success = buildDroid(struct, "Detector Tank", sensorBodies, prop, "", DROID_SENSOR, detectorTurrets);
			}
			break;
		}
		if (success)
			break;
		r+=1;
		if (r>=3)
			r-=3;
	}
	return success;
}

//Produce a tank armed with an EMP weapon.
function produceEMPTank(struct) {
	var emptanks=0;
	for (var gr=0; gr<NUM_GROUPS; ++gr) {
		var ret=groupCountBalance(gr);
		emptanks += ret.emp;
	}
	if (emptanks < MAX_EMP_TANKS) {
		var prop=empPropulsions;
		if (RATE_TANK == 0)
			prop=hoverPropulsions;
		return buildDroid(struct, "EMP Tank", empBodies, prop, "", DROID_WEAPON, empTurrets);
	}
	return false;
}

// produce a cyborg with an anti-personnel weapon in some factory
function produceAPCyborg(struct) {
	if (!BUILD_BORGS)
		return false;
	var iap=random(personality.apCyborgStats.length);
	for (var j=0; j<personality.apCyborgStats[iap].length; ++j) {
		var s = personality.apCyborgStats[iap][j];
		if (buildDroid(struct, "AP Cyborg", s[0], s[2], "", DROID_CYBORG, s[1]))
			return true;
	}
	return false;
}

// produce a cyborg with an anti-tank weapon in some factory
function produceATCyborg(struct) {
	if (!BUILD_BORGS)
		return false;
	var iat=random(personality.atCyborgStats.length);
	for (var j=0; j<personality.atCyborgStats[iat].length; ++j) {
		var s = personality.atCyborgStats[iat][j];
		if (buildDroid(struct, "AT Cyborg", s[0], s[2], "", DROID_CYBORG, s[1]))
			return true;
	}
	return false;
}

// produce a cyborg with an anti-structure weapon in some factory
function produceABCyborg(struct) {
	if (!BUILD_BORGS)
		return false;
	var iap=random(personality.abCyborgStats.length);
	for (var j=0; j<personality.abCyborgStats[iap].length; ++j) {
		var s = personality.abCyborgStats[iap][j];
		if (buildDroid(struct, "AB Cyborg", s[0], s[2], "", DROID_CYBORG, s[1]))
			return true;
	}
	return false;
}

// produce a cyborg with an anti-air weapon in some factory
function produceAACyborg(struct) {
	if (!BUILD_BORGS)
		return false;
	var iap=random(personality.aaCyborgStats.length);
	for (var j=0; j<personality.aaCyborgStats[iap].length; ++j) {
		var s = personality.aaCyborgStats[iap][j];
		if (buildDroid(struct, "AA Cyborg", s[0], s[2], "", DROID_CYBORG, s[1]))
			return true;
	}
	return false;
}

// produce a combat engineer in some factory
function produceEngineer(struct) {
	for (var j=0; j<engineerStats.length; ++j) {
		var s = engineerStats[j];
		if (buildDroid(struct, "Combat Engineer", s[0], s[2], "", DROID_CYBORG_CONSTRUCT, s[1]))
			return true;
	}
	return false;
}

// produce an anti-personnel vtol in some factory
function produceAPVtol(struct) {
	if (!BUILD_VTOLS)
		return false;
	var name = "AP VTOL";
	var prop = vtolPropulsions;
	var body = personality.vtolBodies;
	var weapon1 = personality.vtolWeapons;
	var weapon2 = weapon1;
	var weapon3 = weapon2;
	var iap1=random(weapon1.length);
	var iap2=random(weapon2.length);
	var iap3=random(weapon3.length);
	if (difficulty != EASY) {
		if (random(RATE_VTOL_AP_LIGHT+RATE_VTOL_AP_MEDIUM+RATE_VTOL_AP_HEAVY+RATE_VTOL_AP_SUPER) < RATE_VTOL_AP_SUPER && struct.modules>=2) {
			if (random(RATE_VTOL_AP_SUPER_A+RATE_VTOL_AP_SUPER_B) < RATE_VTOL_AP_SUPER_A) {
				weapon1 = personality.apVtolWeaponsSuper;
				weapon2 = personality.apVtolWeaponsLight;
				iap1=random(weapon1.length);
				iap2=random(weapon2.length);
				body = personality.vtolBodiesSuper;
				name = "Super AP VTOL";
			}
			else {
				weapon1 = personality.apVtolWeaponsHeavy;
				weapon2 = weapon1;
				iap1=random(weapon1.length);
				iap2=random(weapon2.length);
				body = personality.vtolBodiesSuper;
				name = "Super AP VTOL";
			}
		}
		else if (random(RATE_VTOL_AP_LIGHT+RATE_VTOL_AP_MEDIUM+RATE_VTOL_AP_HEAVY) < RATE_VTOL_AP_HEAVY && struct.modules>=2) {
			weapon1 = personality.apVtolWeaponsHeavy;
			iap1=random(weapon1.length);
			body = personality.vtolBodiesHeavy;
			name = "Heavy AP VTOL";
		}
		else if (random(RATE_VTOL_AP_LIGHT+RATE_VTOL_AP_MEDIUM) < RATE_VTOL_AP_MEDIUM && struct.modules>=1) {
			weapon1 = personality.apVtolWeaponsMedium;
			iap1=random(weapon1.length);
			body = personality.vtolBodiesMedium;
			name = "Medium AP VTOL";
		}
		else {
			weapon1 = personality.apVtolWeaponsLight;
			iap1=random(weapon1.length);
			body = personality.vtolBodiesLight;
			name = "Light AP VTOL";
		}
	}
	var ip=random(prop.length);
	var ib=random(body.length);
	return buildDroid(struct, name, body[ib], prop[ip], "", DROID_WEAPON, weapon1[iap1], weapon2[iap2], weapon3[iap3]);
}

// produce an anti-tank vtol in some factory
function produceATVtol(struct) {
	if (!BUILD_VTOLS)
		return false;
	var name = "AT VTOL";
	var prop = vtolPropulsions;
	var body = personality.vtolBodies;
	var weapon1 = personality.vtolWeapons;
	var weapon2 = weapon1;
	var weapon3 = weapon2;
	var iat1=random(weapon1.length);
	var iat2=random(weapon2.length);
	var iat3=random(weapon3.length);
	if (difficulty != EASY) {
		if (random(RATE_VTOL_AT_LIGHT+RATE_VTOL_AT_MEDIUM+RATE_VTOL_AT_HEAVY+RATE_VTOL_AT_SUPER) < RATE_VTOL_AT_SUPER && struct.modules>=2) {
			if (random(RATE_VTOL_AT_SUPER_A+RATE_VTOL_AT_SUPER_B) < RATE_VTOL_AT_SUPER_A) {
				weapon1 = personality.atVtolWeaponsSuper;
				weapon2 = personality.atVtolWeaponsLight;
				iat1=random(weapon1.length);
				iat2=random(weapon2.length);
				body = personality.vtolBodiesSuper;
				name = "Super AT VTOL";
			}
			else {
				weapon1 = personality.atVtolWeaponsHeavy;
				weapon2 = weapon1;
				iat1=random(weapon1.length);
				iat2=random(weapon2.length);
				body = personality.vtolBodiesSuper;
				name = "Super AT VTOL";
			}
		}
		else if (random(RATE_VTOL_AT_LIGHT+RATE_VTOL_AT_MEDIUM+RATE_VTOL_AT_HEAVY) < RATE_VTOL_AT_HEAVY && struct.modules>=2) {
			weapon1 = personality.atVtolWeaponsHeavy;
			iat1=random(weapon1.length);
			body = personality.vtolBodiesHeavy;
			name = "Heavy AT VTOL";
		}
		else if (random(RATE_VTOL_AT_LIGHT+RATE_VTOL_AT_MEDIUM) < RATE_VTOL_AT_MEDIUM && struct.modules>=1) {
			weapon1 = personality.atVtolWeaponsMedium;
			iat1=random(weapon1.length);
			body = personality.vtolBodiesMedium;
			name = "Medium AT VTOL";
		}
		else {
			weapon1 = personality.atVtolWeaponsLight;
			iat1=random(weapon1.length);
			body = personality.vtolBodiesLight;
			name = "Light AT VTOL";
		}
	}
	var ip=random(prop.length);
	var ib=random(body.length);
	return buildDroid(struct, name, body[ib], prop[ip], "", DROID_WEAPON, weapon1[iat1], weapon2[iat2], weapon3[iat3]);
}

// produce an anti-structure vtol in some factory
function produceABVtol(struct) {
	if (!BUILD_VTOLS)
		return false;
	var name = "AB VTOL";
	var prop = vtolPropulsions;
	var body = personality.vtolBodies;
	var weapon1 = personality.vtolWeapons;
	var weapon2 = weapon1;
	var weapon3 = weapon2;
	var iab1=random(weapon1.length);
	var iab2=random(weapon2.length);
	var iab3=random(weapon3.length);
	if (difficulty != EASY) {
		if (random(RATE_VTOL_AB_LIGHT+RATE_VTOL_AB_MEDIUM+RATE_VTOL_AB_HEAVY+RATE_VTOL_AB_SUPER) < RATE_VTOL_AB_SUPER && struct.modules>=2) {
			if (random(RATE_VTOL_AB_SUPER_A+RATE_VTOL_AB_SUPER_B) < RATE_VTOL_AB_SUPER_A) {
				weapon1 = personality.abVtolWeaponsSuper;
				weapon2 = personality.abVtolWeaponsLight;
				iab1=random(weapon1.length);
				iab2=random(weapon2.length);
				body = personality.vtolBodiesSuper;
				name = "Super AB VTOL";
			}
			else {
				weapon1 = personality.abVtolWeaponsHeavy;
				weapon2 = weapon1;
				iab1=random(weapon1.length);
				iab2=random(weapon2.length);
				body = personality.vtolBodiesSuper;
				name = "Super AB VTOL";
			}
		}
		else if (random(RATE_VTOL_AB_LIGHT+RATE_VTOL_AB_MEDIUM+RATE_VTOL_AB_HEAVY) < RATE_VTOL_AB_HEAVY && struct.modules>=2) {
			weapon1 = personality.abVtolWeaponsHeavy;
			iab1=random(weapon1.length);
			body = personality.vtolBodiesHeavy;
			name = "Heavy AB VTOL";
		}
		else if (random(RATE_VTOL_AB_LIGHT+RATE_VTOL_AB_MEDIUM) < RATE_VTOL_AB_MEDIUM && struct.modules>=1) {
			weapon1 = personality.abVtolWeaponsMedium;
			iab1=random(weapon1.length);
			body = personality.vtolBodiesMedium;
			name = "Medium AB VTOL";
		}
		else {
			weapon1 = personality.abVtolWeaponsLight;
			iab1=random(weapon1.length);
			body = personality.vtolBodiesLight;
			name = "Light AB VTOL";
		}
	}
	var ip=random(prop.length);
	var ib=random(body.length);
	return buildDroid(struct, name, body[ib], prop[ip], "", DROID_WEAPON, weapon1[iab1], weapon2[iab2], weapon3[iab3]);
}

// produce an anti-air vtol in some factory
function produceAAVtol(struct) {
	if (!BUILD_VTOLS)
		return false;
	var name = "AA VTOL";
	var prop = vtolPropulsions;
	var body = personality.vtolBodies;
	var weapon1 = personality.vtolWeapons;
	var weapon2 = weapon1;
	var weapon3 = weapon2;
	var iaa1=random(weapon1.length);
	var iaa2=random(weapon2.length);
	var iaa3=random(weapon3.length);
	if (difficulty != EASY) {
		if (random(RATE_VTOL_AA_LIGHT+RATE_VTOL_AA_MEDIUM+RATE_VTOL_AA_HEAVY+RATE_VTOL_AA_SUPER) < RATE_VTOL_AA_SUPER && struct.modules>=2) {
			if (random(RATE_VTOL_AA_SUPER_A+RATE_VTOL_AA_SUPER_B) < RATE_VTOL_AA_SUPER_A) {
				weapon1 = personality.aaVtolWeaponsSuper;
				weapon2 = personality.aaVtolWeaponsLight;
				iaa1=random(weapon1.length);
				iaa2=random(weapon2.length);
				body = personality.vtolBodiesSuper;
				name = "Super AA VTOL";
			}
			else {
				weapon1 = personality.aaVtolWeaponsHeavy;
				weapon2 = weapon1;
				iaa1=random(weapon1.length);
				iaa2=random(weapon2.length);
				body = personality.vtolBodiesSuper;
				name = "Super AA VTOL";
			}
		}
		else if (random(RATE_VTOL_AA_LIGHT+RATE_VTOL_AA_MEDIUM+RATE_VTOL_AA_HEAVY) < RATE_VTOL_AA_HEAVY && struct.modules>=2) {
			weapon1 = personality.aaVtolWeaponsHeavy;
			iaa1=random(weapon1.length);
			body = personality.vtolBodiesHeavy;
			name = "Heavy AA VTOL";
		}
		else if (random(RATE_VTOL_AA_LIGHT+RATE_VTOL_AA_MEDIUM) < RATE_VTOL_AA_MEDIUM && struct.modules>=1) {
			weapon1 = personality.aaVtolWeaponsMedium;
			iaa1=random(weapon1.length);
			body = personality.vtolBodiesMedium;
			name = "Medium AA VTOL";
		}
		else {
			weapon1 = personality.aaVtolWeaponsLight;
			iaa1=random(weapon1.length);
			body = personality.vtolBodiesLight;
			name = "Light AA VTOL";
		}
	}
	var ip=random(prop.length);
	var ib=random(body.length);
	return buildDroid(struct, name, body[ib], prop[ip], "", DROID_WEAPON, weapon1[iaa1], weapon2[iaa2], weapon3[iaa3]);
}

// produce something in fixed template mode
function produceFixedTemplate(struct) {
	var list=[];
	for (var i=0; i<personality.template.length; ++i) {
		if (list.length>=personality.MAX_TEMPLATES)
			break;
		var weapons=1;
		if (!componentAvailable(COMP_BODY,personality.template[i][0]))
			continue;
		if (!componentAvailable(COMP_PROPULSION,personality.template[i][1]))
			continue;
		if (personality.template[i][2] == DROID_WEAPON) {
			if ((personality.template[i][1] == "V-Tol") != (struct.stattype == VTOL_FACTORY))
				continue;
		} else {
			if ((personality.template[i][2] == DROID_CYBORG) != (struct.stattype == CYBORG_FACTORY))
				continue;
		}
		if (!componentAvailable(COMP_WEAPON,personality.template[i][3]))
			continue;
		if (typeof(personality.template[i][4])!="undefined") {
			if (!componentAvailable(COMP_WEAPON,personality.template[i][4])) 
				continue;
			else
				++weapons;
		}
		if (typeof(personality.template[i][5])!="undefined") {
			if (!componentAvailable(COMP_WEAPON,personality.template[i][5])) 
				continue;
			else
				++weapons;
		}
		var n=list.length;
		list[n]=personality.template[i];
		list[n][6]=weapons;
		
	}
	if (list.length <= 0)
		return false;
	var t=list[random(list.length)];
	switch (t[6]) {
	case 1:
		if (buildDroid(struct,"Fixed Template Droid",t[0],t[1],"",t[2],t[3]))
			return true;
		return false;
	case 2:
		if (buildDroid(struct,"Fixed Template Droid",t[0],t[1],"",t[2],t[3],t[4]))
			return true;
		return false;
	case 3:
		if (buildDroid(struct,"Fixed Template Droid",t[0],t[1],"",t[2],t[3],t[4],t[5]))
			return true;
		return false;
	}
}

// get back to work!
function produceDroids() {
	var truckCount=0;
	var factories = enumStruct(me, FACTORY);
	var iHaveFactories=false;
	for (var i=0; i<factories.length; ++i) {
		if (factories[i].status==BUILT) {
			iHaveFactories = true;
			break;
		}
	}
	for (i=0; i<factories.length; ++i) {
		if (outOfPower() && i>0) //Limit production to 1 factory if out of power.
			break;
		struct=factories[i];
		if (structureReady(struct)) {
			//If attacked by Scavenger Choppers, then produce AA droids to ward them off.
			if (attackedByChoppers>0 && iHaveStruct(command) && !personality.fixedTemplates) {
				if (produceAATank(struct))
					continue;
			} 
			if (needMoreTrucks() && truckCount<2) {
				var name = "Truck";
				var body = truckBodies;
				var prop = truckPropulsions;
				var turret = truckTurrets;
				var success = false;
				if (random(3)==0) {
					if (iHaveHovers) {
						name = "Light Hover Truck";
						body = truckBodiesLightHover;
						prop = hoverPropulsions;
						turret = truckTurretsLight;
					}
					else {
						name = "Light Truck";
						body = truckBodiesLightHover;
						prop = truckPropulsions;
						turret = truckTurretsLight;
					}
					success = buildDroid(struct, name, body, prop, "", DROID_CONSTRUCT, turret);
				}
				else {
					switch (struct.modules) {
					case 2:
						name = "Heavy Truck";
						body = truckBodiesHeavy;
						prop = truckPropulsionsHeavy;
						turret = truckTurretsHeavy;
						success = buildDroid(struct, name, body, prop, "", DROID_CONSTRUCT, turret);
						if (success)
							break;
					case 1:
						name = "Medium Truck";
						body = truckBodiesMedium;
						prop = truckPropulsionsMedium;
						turret = truckTurretsMedium;
						success = buildDroid(struct, name, body, prop, "", DROID_CONSTRUCT, turret);
						if (success)
							break;
					case 0:
						name = "Light Truck";
						body = truckBodiesLight;
						prop = truckPropulsionsLight;
						turret = truckTurretsLight;
						success = buildDroid(struct, name, body, prop, "", DROID_CONSTRUCT, turret);
					}
				}
				if (success) {
					++truckCount;
					continue;
				}
			}
			// we shouldn't produce tanks without a cc to avoid template cheating
			if (!iHaveStruct(command) && !personality.fixedTemplates) 
				break;
			if ((!personality.THIS_AI_MAKES_TANKS || random(2)==0) && iHaveArty()) { //Produce sensor tanks when artillery structures become available.
				if (produceSensorTank(struct));
					break;
			}
			/*var abtanks=0, sensortanks=0, cbtanks=0, detectortanks=0;
			for (var gr=0; gr<NUM_GROUPS; ++gr) {
				var ret=groupCountBalance(gr);
				abtanks += ret.ab;
				sensortanks += ret.se;
				cbtanks += ret.cb;
				detectortanks += ret.de;
			}
			if (!personality.THIS_AI_MAKES_TANKS || ((sensortanks*5+1 <= abtanks) && (cbtanks*5+1 <= abtanks) && (detectortanks*5+1 <= abtanks) && enumDroid(me,DROID_WEAPON).length >= personality.MIN_DEFENDERS)) {
				if (produceSensorTank(struct))
					break;
			}*/
			if (!BUILD_TANKS)
				break;
			// we shouldn't produce heavy cannon leopard half-tracks because people will laugh at us
			if (HIGH_TECH_START==1) {
				if (struct.modules==0) {
					continue;
				}
			}
			if (!personality.fixedTemplates) {
				if ((groupSize(aaVtolGroup)+groupSize(aaGroup))<MIN_DEFENDERS_AA) {
					if (produceAATank(struct))
						continue;
				} 
				if (random(5)==0) { //Produce the occasional EMP tank to support our forces.
					if (produceEMPTank(struct))
						break;
				}
				/*if (random(10)==0 || earlyGame(personality.PEACE_TIME)) {
					if (produceHarasser(struct))
						continue;
				}*/
				var j=random(RATE_AP+RATE_AT+RATE_AA+RATE_AB);
				if (j < RATE_AA) {
					if (produceAATank(struct))
						continue;
				} 
				j=random(RATE_AP+RATE_AT+RATE_AB);
				if (j < RATE_AB) {
					if (produceABTank(struct)) 
						continue;
				} 
				j=random(RATE_AP+RATE_AT);
				if (j < RATE_AT)
					if (produceATTank(struct))
						continue;
				if (produceAPTank(struct));
					continue;
			} else 
				produceFixedTemplate(struct);
			//In order for the function to successfully limit which factories produce stuff based on
			//available power, we'll need to queue up and terminate the function here and now,
			//and let structureReady() find the next available factory.
			//queue("produceDroids",100);
			//break;
		}
	}
	var borgfacs = enumStruct(me, CYBORG_FACTORY);
	for (var i=0; i<borgfacs.length; ++i) {
		if (outOfPower() && i>0) //Limit production to 1 factory if out of power.
			return;
		struct=borgfacs[i];
		if (structureReady(struct)) {
			//If attacked by Scavenger Choppers, then produce AA droids to ward them off.
			if (attackedByChoppers>0 && iHaveStruct(command) && !personality.fixedTemplates) {
				if (produceAACyborg(struct))
					continue;
			} 
			if (!personality.fixedTemplates) {
				if (needMoreTrucks() && truckCount<4 && !iHaveFactories) { // don't build borgs when we can build trucks
					produceEngineer(struct);
					++truckCount;
					continue;
				}
			}
			if (!personality.THIS_AI_MAKES_CYBORGS || (RATE_TANK == 0 && status != STATUS_PANIC) || !BUILD_BORGS)
				break;
			if (!personality.fixedTemplates) {
				if ((groupSize(aaVtolGroup)+groupSize(aaGroup))<MIN_DEFENDERS_AA) {
					if (produceAACyborg(struct))
						continue;
				} 
				var j=random(RATE_AP+RATE_AT+RATE_AA+RATE_AB);
				if (j < RATE_AA) {
					if (produceAACyborg(struct))
						continue;
				} 
				j=random(RATE_AP+RATE_AT+RATE_AB);
				if (j < RATE_AB) {
					if (produceABCyborg(struct)) 
						continue;
				} 
				j=random(RATE_AP+RATE_AT);
				if (j < RATE_AT)
					if (produceATCyborg(struct))
						continue;
				produceAPCyborg(struct);
			} else
				produceFixedTemplate(struct);
			//In order for the function to successfully limit which factories produce stuff based on
			//available power, we'll need to queue up and terminate the function here and now,
			//and let structureReady() find the next available factory.
			//queue("produceDroids",100);
			//break;
		}
	}
	var vtolfacs = enumStruct(me, VTOL_FACTORY);
	if (personality.THIS_AI_MAKES_VTOLS) {
		for (var i=0; i<vtolfacs.length; ++i) {
			if (outOfPower() && i>0) //Limit production to 1 factory if out of power.
				return;
			var struct=vtolfacs[i];
			if (structureReady(struct) && BUILD_VTOLS) {
				if (!personality.fixedTemplates) {
					if ((groupSize(aaVtolGroup)+groupSize(aaGroup))<MIN_DEFENDERS_AA) {
						if (produceAAVtol(struct))
							continue;
					} 
					var j=random(RATE_AP+RATE_AT+RATE_AA+RATE_AB);
					if (j < RATE_AA) {
						if (produceAAVtol(struct))
							continue;
					} 
					j=random(RATE_AP+RATE_AT+RATE_AB);
					if (j < RATE_AB) {
						if (produceABVtol(struct)) 
							continue;
					} 
					j=random(RATE_AP+RATE_AT);
					if (j < RATE_AT)
						if (produceATVtol(struct))
							continue;
					produceAPVtol(struct);
				} else
					produceFixedTemplate(struct);
				//In order for the function to successfully limit which factories produce stuff based on
				//available power, we'll need to queue up and terminate the function here and now,
				//and let structureReady() find the next available factory.
				//queue("produceDroids",100);
				//break;
			}
		}
	}
}


/* Attacking the enemy */


// checks if a player was not defeated yet
function isAlive(player) {
	for (var i=0; i<allInAttackTargets.length; ++i)
		if (enumStruct(player, allInAttackTargets[i]).length>0)
			return true;
	if (enumDroid(player).length>0)
		return true;
	return false;
}

// checks if the enemy is actually an threat
function isAnEnemy(enemy) {
	if (typeof(enemy)=="undefined")
		return false;
	if (enemy < 0 || enemy >= maxPlayers) {
		// probably scavengers ...
		return false;
	}
	if (allianceExistsBetween(me, enemy)) 
		return false;
	if (isAlive(enemy))
		return true;
	return false;
}

// cycle through enemies
function nextEnemy(enemy) {
	var i=enemy;
	do {
		++i;
		if (i>maxPlayers) 
			i=0;
		if (isAnEnemy(i) && enemyInfo[i].reachable)
			return i;
	} while (i!=enemy);
	do {
		++i;
		if (i>maxPlayers) 
			i=0;
		if (isAnEnemy(i))
			return i;
	} while (i!=enemy);
	return i;
}

// how many enemies left
function countEnemies() {
	var ret=0;
	for (var i=0; i<maxPlayers; ++i)
		if (isAnEnemy(i))
			++ret;
	return ret;
}

// hom many allies do i have
function countAllies() {
	var ret=0;
	for (var i=0; i<maxPlayers; ++i)
		if (allianceExistsBetween(me,i))
			if (isAlive(i))
				++ret;
	return ret;
	
}

// how many teams the enemies are split into
function countEnemyTeams() {
	var forces=[];
	for (var i=0; i<maxPlayers; ++i) if (isAnEnemy(i)) {
		var found=false;
		for (var j=0; j<forces.length; ++j) 
			if (allianceExistsBetween(i,forces[j])) {
				found=true;
				break;
			}
		if (!found)
			forces[forces.length]=i;
	}
	return forces.length;
}

// picks an enemy player to attack
function findEnemy() {
	var enemies=[];
	for (var i=0; i<maxPlayers; ++i)
		if (isAnEnemy(i) && enemyInfo[i].reachable) // prefer ground enemies
			enemies[enemies.length]=i;
	if (enemies.length>0)
		return enemies[random(enemies.length)];
	enemies=[];
	for (var i=0; i<maxPlayers; ++i)
		if (isAnEnemy(i))
			enemies[enemies.length]=i;
	return enemies[random(enemies.length)];
	
}

// Checks if we're in a position to launch attacks on a certain enemy's home base.
// Note that, depending on the AI's PEACEFULNESS value, the AI may sometimes launch full-scale assaults on enemies when outnumbered!
function okToAllIn(enemy) {
	if (!isAlive(enemy) || earlyGame(personality.PEACE_TIME))
		return false; // this is rather a clean-up, we need to look for derricks *again* when the opponent is almost dead
	else {
		if (enumDroid(me).length >= ALL_IN_THRESHOLD) //This check is there in case the AI is close to not being able to build any more units.
			return true;
		var goodGuys=0, badGuys=0;
		badGuys += enumDroid(enemy, DROID_WEAPON).length;
		badGuys += (enumDroid(enemy, DROID_CYBORG).length)/4; //Heavy Cyborgs have roughly 1/4 of the HP, price, and build points of light tanks in Contingency.
		var enemyStructs=enumStruct(enemy,DEFENSE);
		for (var i=0; i<enemyStructs.length; ++i)
			if (!isDedicatedAA(enemyStructs[i]))
				badGuys += .25; //Treat defenses as 1/4 of a unit since they can't move around (obviously).
		/*for (var i=0; i<NUM_GROUPS; ++i) {
			if (groupInfo[i].enemy==enemy)
				goodGuys+=groupSize(battleGroup[i]);
		}*/
		goodGuys += enumDroid(me, DROID_WEAPON).length;
		goodGuys += (enumDroid(me, DROID_CYBORG).length)/4;
		if (PEACEFULNESS * badGuys < goodGuys) {
			return true;
		}
	}
	return false;
}

// finds an appropriate target within enemy (enemy=-1 to choose random enemy)
function findTarget(enemy) {
	if (enemy==-1)
		enemy=findEnemy();
	if (typeof(enemy)=="undefined")
		return;
	var list=allInAttackTargets;
	if (scavengerPlayer!=-1) {
		if (okToAllIn(enemy)) //We should only assault Scavenger bases if we have a lot of forces.
			list=list.concat(enumStructArray(scavengerPlayer, [derrick,FACTORY,VTOL_FACTORY,]));
		else { //Otherwise, we should avoid them, as they can often defend their derricks reasonably well.
			list=attackTargets;
			for (var i=0; i<maxPlayers; ++i)
				if (isAnEnemy(i))
					list=list.concat(enumStruct(i, derrick)); //We should take out any enemy derricks (though not Scavenger-controlled ones) that may be around.
		}
	}
	var obj=undefined, danger=Infinity, distance=Infinity;
	for (var i=list.length-1; i>=0; --i) {
		var structs=enumStruct(enemy, list[i]);
		if (structs.length > 0) {
			for (var j=0; j<structs.length; ++j) {
				var _danger = dangerLevel(structs[j].x,structs[j].y)
				var _distance = distBetweenTwoPoints(basePosition.x,basePosition.y,structs[j].x,structs[j].y);
				if ((_danger<danger) || (_danger==danger && _distance<distance)) {
					danger=_danger;
					distance=_distance;
					obj=structs[j];
				}
			}
		}
	}
	if (typeof(obj)!="undefined")
		return obj;
	var droids=enumDroid(enemy);
	if (droids.length > 0) {
		return droids[random(droids.length)];
	}
}

// checks if the target set for a group is already dead
// and updates if necessary
function getGroupTarget(gr) {
	if (groupInfo[gr].helpPlayer!=-1) {
		if (gameTime - groupInfo[gr].helpLastUpdate > 300000) {
			// stop helping after 5 minutes
			groupInfo[gr].helpPlayer = -1;
		} else {
			groupInfo[gr].targetX = groupInfo[gr].helpX;
			groupInfo[gr].targetY = groupInfo[gr].helpY;
			groupInfo[gr].targetType = -1;
			return {x:groupInfo[gr].targetX,y:groupInfo[gr].targetY};
		}
	}
	if (groupInfo[gr].targetAge>0) { // if doesn't need updating yet
		// check if the currently selected target is valid
		var x=groupInfo[gr].targetX;
		var y=groupInfo[gr].targetY;
		if (x!=-1 && y!=-1) {
			if (groupInfo[gr].targetType != TARGET_DROID) {
				var list=enumStruct(groupInfo[gr].enemy,groupInfo[gr].targetStatType);
				for (var k=0; k<list.length; ++k)
					if (list[k].x==x && list[k].y==y) {
						return list[k];
					}
			} else {
				var list=enumDroid(groupInfo[gr].enemy,groupInfo[gr].targetStatType);
				for (var k=0; k<list.length; ++k)
					if (distBetweenTwoPoints(x,y,list[k].x,list[k].y==y)<3) {
						return list[k];
					}
			}
		}
	}
	// pick a new one otherwise
	if (!isAnEnemy(groupInfo[gr].enemy))
		groupInfo[gr].enemy=findEnemy();
	var target=findTarget(groupInfo[gr].enemy);
	if (typeof(target)!="undefined") {
		groupInfo[gr].targetX = target.x;
		groupInfo[gr].targetY = target.y;
		groupInfo[gr].targetType = TARGET_STRUCTURE;
		groupInfo[gr].targetAge = 20; // around 3 minutes before target update
		if (target.type == DROID) {
			groupInfo[gr].targetType = TARGET_DROID;
			groupInfo[gr].targetStatType = target.droidType;
		} else
			groupInfo[gr].targetStatType = target.stattype;
		return target;
	}
	
}

// a very naive way of finding where our units mostly are
function naiveFindClusters(list,size) {
	this.clusters=[];
	this.xav=[];
	this.yav=[];
	this.maxIdx=0;
	this.maxCount=0;
	for (i=list.length-1; i>=0; --i) {
		var x=list[i].x,y=list[i].y;
		var found=false;
		for (var j=0; j<this.clusters.length; ++j) {
			if (distBetweenTwoPoints(this.xav[j],this.yav[j],x,y)<size) {
				var n=this.clusters[j].length;
				this.clusters[j][n]=i;
				this.xav[j]=(n*this.xav[j]+x)/(n+1);
				this.yav[j]=(n*this.yav[j]+y)/(n+1);
				if (this.clusters[j].length>this.maxCount) {
					this.maxIdx=j;
					this.maxCount=this.clusters[j].length;
				}
				found=true;
				break;
			}
		}
		if (!found) {
			var n=this.clusters.length;
			this.clusters[n]=[i];
			this.xav[n]=x;
			this.yav[n]=y;
			if (1>this.maxCount) {
				this.maxIdx=n;
				this.maxCount=1
			}
		}
	}
}

// order a droid to stop moving
function stopDroid(droid) {
	if (typeof(orderDroid)=="undefined")
		orderDroidLoc(droid, DORDER_MOVE, droid.x, droid.y);
	else
		orderDroid(droid, DORDER_STOP);
}

// for formation movement of groups
// returns false iff already in formation
function regroupWarriors(gr) {
	var list=enumGroup(battleGroup[gr]);
	if (DEF_LIGHT==1) {
		if (groupSize(battleGroup[gr])<MIN_WARRIORS) {
			for (var i=0; i<list.length; ++i)
				if (!sendForRepair(list[i])) {
					//orderDroidLoc(list[i], DORDER_MOVE, list[ret.maxIdx].x,list[ret.maxIdx].y);
					if (groupInfo[gr].delaystaticregroup > 0)
						returnToBase(list[i]);
					else
						stopDroid(list[i]);
				}
		}
		return groupSize(battleGroup[gr])<MIN_WARRIORS;
	}
	if (typeof(regroupWarriors.lastUnstuck)==undefined || isNaN(regroupWarriors.lastUnstuck))
		regroupWarriors.lastUnstuck=maxPlayers;
	if (typeof(regroupWarriors.lastReturn)==undefined)
		regroupWarriors.lastReturn=false;
	// sometimes groups have a right to be dispersed
	if (groupInfo[gr].delayregroup>0)
		return regroupWarriors.lastReturn;
	// don't bother doing this sort of thing in case of emergency
	if (status!=STATUS_NORMAL)
		return false;
	// turtle AI doesn't really need to regroup his sensors
	if (!personality.THIS_AI_MAKES_TANKS && !personality.THIS_AI_MAKES_CYBORGS)
		return false;
	// early rushy tanks need to be fast
	if (earlyGame(personality.PEACE_TIME))
		return false;
	var min = groupSize(battleGroup[gr])*0.6;
	if (this.delaystaticregroup > 0)
		min = groupSize(battleGroup[gr])*0.4;
	if (min < MIN_WARRIORS)
		min = MIN_WARRIORS;
	var ret=new naiveFindClusters(list,MAX_DISPERSE + list.length/12);
	if (ret.maxCount >= min) {
		if (distBetweenTwoPoints(ret.xav[ret.maxIdx],ret.yav[ret.maxIdx],groupInfo[gr].lastX,groupInfo[gr].lastY)<5) {
			++groupInfo[gr].idleTime;
		} else {
			groupInfo[gr].idleTime=0;
		}
	} else {
		groupInfo[gr].idleTime=0;
	}
	groupInfo[gr].lastX=ret.xav[ret.maxIdx];
	groupInfo[gr].lastY=ret.yav[ret.maxIdx];
	if (groupInfo[gr].idleTime<3) {
		if (
			(groupInfo[gr].delaystaticregroup == 0) 
				|| 
			(ret.maxCount > min)
		) {
			for (var i=0; i<ret.clusters.length; ++i) 
				if (i!=ret.maxIdx)  {
					for (j=0; j<ret.clusters[i].length; ++j) 
						if (!sendForRepair(list[ret.clusters[i][j]]) && !droidIsAtHome(list[ret.clusters[i][j]]))
							orderDroidLoc(list[ret.clusters[i][j]], DORDER_MOVE, ret.xav[ret.maxIdx], ret.yav[ret.maxIdx]);
				} else if (ret.maxCount < min || ret.maxCount < list.length/2) { // note: i==ret.maxIdx here
					for (j=0; j<ret.clusters[i].length; ++j) 
						if (!sendForRepair(list[ret.clusters[i][j]]) && !droidIsAtHome(list[ret.clusters[i][j]]))
							stopDroid(list[ret.clusters[i][j]]);
				}
		} else {
			for (var i=0; i<list.length; ++i)
				if (!sendForRepair(list[i]) && !droidIsAtHome(list[i])) {
					//orderDroidLoc(list[i], DORDER_MOVE, list[ret.maxIdx].x,list[ret.maxIdx].y);
					if (groupInfo[gr].delaystaticregroup > 0)
						returnToBase(list[i]);
					else
						stopDroid(list[i]);
				}
		}
		groupInfo[gr].delayregroup = 1; // don't regroup too often
		regroupWarriors.lastReturn = (ret.maxCount < min);
		return regroupWarriors.lastReturn;
	} else if (countEnemies()>0) { // anti-stuck mechanism below
		var rx,ry;
		++regroupWarriors.lastUnstuck;
		if (regroupWarriors.lastUnstuck>=maxPlayers) {
			regroupWarriors.lastUnstuck=0;
			rx=random(mapWidth-2)+1;
			ry=random(mapHeight-2)+1;
		} else {
			rx=startPositions[regroupWarriors.lastUnstuck].x;
			ry=startPositions[regroupWarriors.lastUnstuck].y;
		}
		list=enumDroid(me); //ANY unit, not just combat units, can get themselves stuck, so we'll need to order EVERY single unit (except for VTOLs) to the same location to get them unstuck.
		for (var i=0; i<list.length; ++i) {
			if (isVTOL(list[i])) {
				if (list[i].droidType==DROID_WEAPON || list[i].droidType==DROID_CYBORG)
					orderDroidLoc(list[i], DORDER_SCOUT, rx, ry);
				else
					orderDroidLoc(list[i], DORDER_MOVE, rx, ry);
			}
		}
		/*list=enumDroid(me,DROID_WEAPON);
		list=list.concat(enumDroid(me,DROID_CYBORG));
		for (var i=0; i<list.length; ++i)
			orderDroidLoc(list[i], DORDER_SCOUT, rx, ry);*/
		for (var i=0; i<NUM_GROUPS; ++i)
			groupInfo[i].delayregroup = 4; // for around half a minute, shouldn't be enough to reach too far
		setStatus(STATUS_UNSTUCK);
		regroupWarriors.lastReturn = true;
		return regroupWarriors.lastReturn;
	}
}

function regroupRegularly() {
	if (typeof(regroupRegularly.lastGroup)=="undefined")
		regroupRegularly.lastGroup = 0;
	regroupWarriors(regroupRegularly.lastGroup);
	++regroupRegularly.lastGroup;
	if (regroupRegularly.lastGroup>=NUM_GROUPS)
		regroupRegularly.lastGroup = 0;
}

function findNearestTarget(x,y) {
	var minDist=Infinity, minObj;
	for (var i=0; i<maxPlayers; ++i) if (isAnEnemy(i)) {
		var list=enumStruct(i);
		if (personality.DEFENSIVENESS==0)
			list=list.concat(enumDroid(i,DROID_WEAPON));
		for (var j=0; j<list.length; ++j) {
			var dist=distBetweenTwoPoints(list[j].x,list[j].y,x,y);
			if (dist<minDist) {
				minDist = dist;
				minObj = list[j];
			}
		}
	}
	return minObj;
}

// find the next non-empty group
function nextGroup(gr) {
	var next=gr;
	while (true) {
		++next;
		if (next>=NUM_GROUPS)
			next=0;
		if (next==gr)
			return gr;
		if (groupSize(battleGroup[next])>0)
			return next;
	}
}

// send harassers to knock out defenseless targets
/*function harassStuff() {
	var droids=enumGroup(harassGroup);
	var freeDroids=0;
	for (var k=0; k<droids.length; ++k) 
		if (droids[k].order!=DORDER_SCOUT)
			++freeDroids;
	if (freeDroids==0)
		return;
	//var list=cachedEnumEnemies(findEnemy());
	var list=enumStruct(scavengerPlayer, derrick);
	for (var i=0; i<maxPlayers; ++i)
		if (i!=me && !allianceExistsBetween(me,i))
			list=list.concat(enumStruct(i,derrick));
	for (i=0; i<10; ++i) {
		j=random(list.length);
		if (typeof(list[j])!="undefined" && typeof(list[j].x)!="undefined" && typeof(list[j].y)!="undefined") {
			if (safeDest(me,list[j].x,list[j].y)) {
				for (var k=0; k<droids.length; ++k) if (droids[k].order!=DORDER_SCOUT) {
					orderDroidLoc(droids[k],DORDER_SCOUT,list[j].x,list[j].y);
				}
			}
		}
	}
}*/

// send harassers to knock out defenseless targets
function harassStuff() {
	var droids=enumGroup(harassGroup);
	var freeDroids=0;
	for (var k=0; k<droids.length; ++k) 
		if (droids[k].order!=DORDER_SCOUT)
			++freeDroids;
	if (freeDroids==0)
		return;
	//var list=cachedEnumEnemies(findEnemy());
	var list=enumStruct(scavengerPlayer, derrick);
	var enemies=[];
	for (var i=0; i<maxPlayers; ++i) {
		if (i!=me && !allianceExistsBetween(me,i)) {
			var derricks=enumStruct(i,derrick);
			if (isAlive(i))
				enemies[enemies.length]=i;
			if (!isAlive(i) || difficulty<=MEDIUM) //MEDIUM AIs may attempt to rush an enemy's base with their harassers.
				list=list.concat(derricks);
			else if (derricks.length>0) { //HARD and INSANE AIs, on the other hand, should keep their harassers away from enemy bases, especially during the start of a game.
				for (var j=0; j<derricks.length; ++j) {
					if (distBetweenTwoPoints(derricks[j].x,derricks[j].y,startPositions[i].x,startPositions[i].y)>BASE_SIZE)
						list[list.length]=derricks[j];
				}
			}
		}
	}
	var success=false;
	if (list.length>0) {
		for (var j=0; j<droids.length; ++j) {
			if (droids[j].order!=DORDER_SCOUT) {
				list=sortByDistance(droids[j].x,droids[j].y,list);
				for (i=0; i<list.length; ++i) {
					if (random(3)==0) {
						orderDroidLoc(droids[j],DORDER_SCOUT,list[i].x,list[i].y);
						success=true;
						break;
					}
				}
			}
		}
		if (success)
			return true;
	} //If no suitable derricks are found, then order the harassers to guard the various oil resources outside of enemy bases.
	list=enumFeature(-1, oilres);
	if (list.length>0 && enemies.length>0) {
		for (i=0; i<10; ++i) {
			var r=enemies[random(enemies.length)];
			list=sortByDistance(startPositions[r].x,startPositions[r].y,list);
			for (var j=0; j<list.length; ++j) {
				var success2=true;
				for (var k=0; k<enemies.length; ++k) {
					if (distBetweenTwoPoints(list[j].x,list[j].y,startPositions[enemies[k]].x,startPositions[enemies[k]].y)<BASE_SIZE) {
						success2=false;
						break;
					}
				}
				if (success2 || difficulty<=MEDIUM) { //MEDIUM AIs will most likely rush into the enemy's base.
					for (var k=0; k<droids.length; ++k) if (droids[k].order!=DORDER_SCOUT) {
						if (random(3)!=0) {
							orderDroidLoc(droids[k],DORDER_SCOUT,list[j].x,list[j].y);
						}
					}
				}
			}
		}
	}
}

// the attack
function attackStuff() {
	if (typeof(DORDER_OBSERVE)=="undefined")
		var DORDER_OBSERVE = 9; // HACK: waiting until this constant is exported to scripts ...
	if (typeof(attackStuff.groupNumber)=="undefined")
		attackStuff.groupNumber = 0;
	attackStuff.groupNumber = nextGroup(attackStuff.groupNumber);
	// don't bother doing this sort of thing in case of emergency
	if (status==STATUS_PANIC && difficulty<HARD)
		return;
	var sensorAttack = (!personality.THIS_AI_MAKES_TANKS && !personality.THIS_AI_MAKES_CYBORGS);
	var sensorTarget;
	var gr=attackStuff.groupNumber;
	var attackers=enumGroup(battleGroup[gr]);
	if (sensorAttack || groupSize(battleGroup[gr]) >= MIN_WARRIORS || (groupSize(battleGroup[gr]) >= NO_WARRIORS_AT_ALL && earlyGame(personality.PEACE_TIME))) {
		var needRegrouping = regroupWarriors(gr);
		var target=getGroupTarget(gr);
		if (typeof(target) != "undefined") {
			for (var i=0; i<attackers.length; ++i) {
				droid=attackers[i];
				if (needRegrouping && !droidIsAtHome(droid))
					continue;
				if (sendForRepair(droid))
					continue;
				if (!droidCanReach(droid,target.x,target.y)) { // useless droid ...
					groupAddDroid(defendGroup,droid);
					continue;
				}
				if (droid.order != DORDER_ATTACK && droid.order != DORDER_SCOUT) {
					if (droid.name.indexOf("Sensor")> -1) {
						if (typeof(sensorTarget)!="undefined")
							sensorTarget=findNearestTarget(droid.x,droid.y);
						if (typeof(sensorTarget)!="undefined")
							orderDroidObj(droid, DORDER_OBSERVE, sensorTarget);
						else if (groupInfo[gr].targetType!=-1)
							orderDroidObj(droid, DORDER_OBSERVE, target);
					} else orderDroidLoc(droid, DORDER_SCOUT, target.x+random(2)-1, target.y+random(2)-1);
				}
			}
		}
	}
}

// attack things with VTOLs
function vtolAttack() {
	if (!personality.THIS_AI_MAKES_VTOLS)
		return;
	//Order ap and at vtols to attack the closest enemy units.
	for (var gr=0; gr<NUM_VTOL_GROUPS; ++gr) if (groupSize(vtolGroup[gr])>=MIN_VTOLS) {
		var patrol=1;
		var targets = sortByDistance(basePosition.x,basePosition.y,cachedEnumEnemies());
		if (targets.length>0) { //Check for enemy droids 1st.
			for (var i=0; i<targets.length; ++i) {
				if (targets[i].type==DROID) {
					if (!isVTOL(targets[i])) {
						if (random(3)==0) {
							var target=targets[i];
							break;
						}
					}
				}
			} //If no enemy droids are available, then look for structures.
			if (typeof(target) == "undefined") {
				for (var i=0; i<targets.length; ++i) {
					if (targets[i].type==STRUCTURE) {
						if (random(3)==0) {
							var target=targets[i];
							break;
						}
					}
				}
			} //Still no luck? Then attack the closest enemy, regardless of what it is.
			if (typeof(target) == "undefined") 
				var target=targets[0];
			if (typeof(target) != "undefined") {
				var vtols=enumGroup(vtolGroup[gr]);
				var list=[];
				for (var i=0; i<vtols.length; ++i) {
					var vtol=vtols[i];
					if (isNaN(vtol.armed)) {
						if (vtolAttack.debugSent!=true) {
							debug("vtol.armed is NaN, assuming fully rearmed - known v3.1 bug, see http://developer.wz2100.net/ticket/3178");
							vtolAttack.debugSent=true;
						}
					} 
					if (vtol.armed >= 99 || isNaN(vtol.armed)) {
						if (vtol.order != DORDER_ATTACK)
							list[list.length]=i;
					}
				}
				if (list.length>=MIN_VTOLS) {
					for (var i=0; i<list.length; ++i) {
						if (patrol==0)
							orderDroidObj(vtols[list[i]], DORDER_ATTACK, target);
						else
							orderDroidLoc(vtols[list[i]], DORDER_PATROL, target.x, target.y);
					}
				}
				else if (list.length>0) //Until all vtols are ready, order those that are to protect the main base.
					for (var i=0; i<list.length; ++i)
						orderDroidLoc(list[i], DORDER_CIRCLE, basePosition.x, basePosition.y);
			}
		}
	}
	//Order ab vtols to attack enemy base structures.
	if (groupSize(abVtolGroup)>=MIN_VTOLS) {
		var patrol=0;
		var targets = sortByDistance(basePosition.x,basePosition.y,cachedEnumEnemies());
		if (targets.length>0) {		//Sometimes, we should look for enemy base structures 1st, starting with those that are located 
			if (random(2)==0) {	//at the bottom of the allInAttackTargets array (which are classified as higher-priority base structures).
				for (var j=allInAttackTargets.length-1; j>=0; --j) {
					for (var i=0; i<targets.length; ++i) {
						if (targets[i].name==allInAttackTargets[j]) {
							if (random(2)==0) {
								var abTarget=targets[i];
								break;
							}
						}
					}
					if (typeof(abTarget) != "undefined")
						break;
				}
			} //At other times, we should just attack the closest structure defined in the allInAttackTargets array.
			for (var i=0; i<targets.length; ++i) {
				for (var j=allInAttackTargets.length-1; j>=0; --j) {
					if (targets[i].name==allInAttackTargets[j]) {
						if (random(2)==0) {
							var abTarget=targets[i];
							break;
						}
					}
				}
				if (typeof(abTarget) != "undefined")
					break;
			} //If no suitable base structures are available, then look for any kind of structure.
			if (typeof(abTarget) == "undefined") {
				for (var i=0; i<targets.length; ++i) {
					if (targets[i].type==STRUCTURE) {
						if (random(3)==0) {
							var abTarget=targets[i];
							patrol=1;
							break;
						}
					}
				}
			} //Still no luck? Then attack the closest enemy, regardless of what it is.
			if (typeof(abTarget) == "undefined") {
				var abTarget=targets[0];
				patrol=1;
			}
			if (typeof(abTarget) != "undefined") {
				var vtols=enumGroup(abVtolGroup);
				var list=[];
				for (var i=0; i<vtols.length; ++i) {
					var vtol=vtols[i];
					if (isNaN(vtol.armed)) {
						if (vtolAttack.debugSent!=true) {
							debug("vtol.armed is NaN, assuming fully rearmed - known v3.1 bug, see http://developer.wz2100.net/ticket/3178");
							vtolAttack.debugSent=true;
						}
					} 
					if (vtol.armed >= 99 || isNaN(vtol.armed)) {
						if (vtol.order != DORDER_ATTACK)
							list[list.length]=i;
					}
				}
				if (list.length>=MIN_VTOLS) {
					for (var i=0; i<list.length; ++i) {
						if (patrol==0)
							orderDroidObj(vtols[list[i]], DORDER_ATTACK, abTarget);
						else
							orderDroidLoc(vtols[list[i]], DORDER_PATROL, abTarget.x, abTarget.y);
					}
				}
				else if (list.length>0) { //Send some of our ab VTOLs after some enemy derricks.
					for (var i=0; i<list.length; ++i)
						for (var j=0; j<targets.length; ++j)
							if (targets[j].stattype==RESOURCE_EXTRACTOR)
								if (distBetweenTwoPoints(targets[j].x,targets[j].y,startPositions[targets[j].player].x,startPositions[targets[j].player].y)>BASE_SIZE)
									if (random(3)==0)
										orderDroidObj(vtols[list[i]], DORDER_ATTACK, targets[j]);
				}
			}
		} 
	}
	//Order aa vtols to defend the base.
	var vtols=enumGroup(aaVtolGroup);
	for (var i=0; i<vtols.length; ++i) {
		var vtol=vtols[i];
		if (isNaN(vtol.armed)) {
			if (vtolAttack.debugSent!=true) {
				debug("vtol.armed is NaN, assuming fully rearmed - known v3.1 bug, see http://developer.wz2100.net/ticket/3178");
				vtolAttack.debugSent=true;
			}
		} 
		if (vtol.armed >= 99 || isNaN(vtol.armed)) {
			if (vtol.order != DORDER_ATTACK && vtol.order != DORDER_CIRCLE)
				orderDroidLoc(vtol, DORDER_CIRCLE, basePosition.x, basePosition.y);
		}
	}
}

// make sure the defenders of the base, as well as the trucks that are in charge of base construction/maintenance, are somewhere near the base
function returnDefendersToBase() {
	var list=enumGroup(defendGroup).concat(enumGroup(aaGroup),enumGroup(builderGroup));
	for (var i=0; i<list.length; ++i)
		if (!isTruckBusy(list[i]))
			returnToBase(list[i]);
}

function callUpdateEnemyInfo() {
	var noReturn=(typeof(enemyInfo)=="undefined");  // this usually happens on loading a game, need a full update in that case
	for (var i=-1; i<maxPlayers; ++i) {
		if (typeof(enemyInfo[i])=="undefined")
			noReturn=true;
	}
	if (noReturn) {
		global.enemyInfo = [];
		enemyInfo[-1] = new Object();
	}
	for (var i=0; i<maxPlayers; ++i) if (isAnEnemy(i)) {
		if (typeof(enemyInfo[i])=="undefined" || gameTime - enemyInfo[i].lastUpdate > 30000) {
			enemyCountBalance(i);
			if (!noReturn) return;
		}
	}
	if (typeof(enemyInfo[-1])=="undefined" || gameTime - enemyInfo[-1].lastUpdate > 30000) {
		enemyInfo[-1].tank = 0;
		enemyInfo[-1].borg = 0;
		enemyInfo[-1].vtol = 0;
		enemyInfo[-1].defs = 0;
		enemyInfo[-1].aa = 0;
		for (var i=0; i<maxPlayers; ++i) if (isAnEnemy(i)) {
			enemyInfo[-1].tank += enemyInfo[i].tank;
			enemyInfo[-1].borg += enemyInfo[i].borg;
			enemyInfo[-1].vtol += enemyInfo[i].vtol;
			enemyInfo[-1].defs += enemyInfo[i].defs;
			enemyInfo[-1].aa   += enemyInfo[i].aa;
		}
		enemyInfo[-1].lastUpdate = gameTime;
	}
}

function fallBack(droid,threat) {
	if (droid.order==DORDER_MOVE || difficulty==EASY) //EASY AIs shouldn't be capable of micromanaging their own units.
		return;
	var x=droid.x-3*(threat.x-droid.x)/distBetweenTwoPoints(threat.x,threat.y,droid.x,droid.y);
	var y=droid.y-3*(threat.y-droid.y)/distBetweenTwoPoints(threat.x,threat.y,droid.x,droid.y);
	if (x<1 || y<1 || x>mapWidth-2 || y>mapHeight-2)
		return;
	orderDroidLoc(droid,DORDER_MOVE,x,y);
}

	/**********/
	/* Events */
	/**********/
	
// work harder
function eventDroidIdle(droid) {
	if (droid.droidType == DROID_CONSTRUCT || droid.droidType == DROID_CYBORG_CONSTRUCT) {
		queue("huntForOil");
		queue("buildDefenses");
		queue("buildLookoutTower");
		queue("buildArtillery");
		queue("buildAADefense");
		queue("buildRepairFacility");
		queue("buildProxyFactory");
	}
}

// something was just researched
function eventResearched(tech, labparam) {
	queue("doResearch");
	queue("executeBuildOrder");
	queue("produceDroids");
}

// something was just built
function eventStructureBuilt(struct, droid) {
	if (struct.stattype == RESOURCE_EXTRACTOR) {
		++derrickStats[findDerrickStat(struct)].hotness;
	}
	queue("executeBuildOrder", 200);
	queue("huntForOil");
	queue("produceDroids");
	queue("doResearch");
}

// something was just produced
function eventDroidBuilt(droid, struct) {
	if (droid.player != me)
		return;
	if (isVTOL(droid))
		addVtolToSomeGroup(droid);
	else if (droid.droidType == DROID_CONSTRUCT) {
			addTruckToSomeGroup(droid);
			queue("executeBuildOrder");
			queue("huntForOil");
	} else {
		if (tankReachibilityCheck(droid))
			RATE_TANK=0;
		addTankToSomeGroup(droid);
	}
	queue("produceDroids");
}

// something was just attacked
function eventAttacked(victim, attacker) {
	if (typeof(victim)=="undefined" || typeof(attacker)=="undefined") 
		return;
	if (typeof(eventAttacked.lastCall)!="undefined" && gameTime - eventAttacked.lastCall < 200) {
		if (victim.type == DROID && !isVTOL(victim))
			fallBack(victim,attacker);
		return;
	}
	if (typeof(attacker.player)!="undefined") 
		if (attacker.player >= 0 && attacker.player < maxPlayers)
			if (allianceExistsBetween(attacker.player,me)) // don't respond to friendly splash damage
				return;
	eventAttacked.lastCall = gameTime;
	personality.PEACE_TIME = 0;
	if (status==STATUS_EARLYGAME)
		setStatus(STATUS_NORMAL);
	var dist=distBetweenTwoPoints(attacker.x,attacker.y,victim.x,victim.y);
	var arty=false;
	if (typeof(attacker.hasIndirect) == "undefined") {
		if (dist>ARTILLERY_RANGE)
			arty=true;
	} else {
		if (attacker.hasIndirect && dist>ARTILLERY_RANGE) //We need to ignore short-range artillery such as Mortars and Mini-Rocket Arrays.
			arty=true;
	}
	var vtol=false;
	if (attacker.type==DROID)
		if (isVTOL(attacker))
			vtol=true;
	if (!arty && !vtol) //Don't create hotspots near targets of enemy counter-battery and counter-radar fire, or targets of enemy VTOL attack runs.
		createHotspot(victim.x,victim.y);
	if (attacker.type == DROID) 
		if (isVTOL(attacker) || attacker.droidType==DROID_TRANSPORTER || attacker.droidType==DROID_SUPERTRANSPORTER) {
			if (!warnVtols) {
				if (typeof(chat)!="undefined") {
					if (attacker.player!=scavengerPlayer)
						chat(ALLIES,"Player "+attacker.player+" is using VTOLs!")
					else
						chat(ALLIES,"Scavengers are using VTOLs!")
				}
				//if (MIN_DEFENSES_AA<8)
					//MIN_DEFENSES_AA=8; //Start building some anti-air defenses within our base.
				if (MIN_DEFENDERS_AA<16)
					MIN_DEFENDERS_AA=16; //Start producing some anti-air units to protect our base.
				warnVtols=true;
			}
			// Build anti-air defenses at spots of enemy VTOL attacks
			lastAirAttackX = victim.x;
			lastAirAttackY = victim.y;
			return;
		}
	if (victim.type == DROID) {
		if (isVTOL(victim)) {
			// VTOLs fire back when attacked by AA
			if (!needsRepair(victim))
				if (victim.health > RETREAT_AT)
					orderDroidObj(victim, DORDER_ATTACK, attacker);
			return;
		}
		if (victim.droidType == DROID_WEAPON || victim.droidType == DROID_CYBORG || victim.droidType == DROID_SENSOR) {
			if (victim.group==harassGroup) {
				if (!needsRepair(victim)) {
					if (victim.health > RETREAT_AT) {
						orderDroidObj(victim, DORDER_ATTACK, attacker);
						return;
					}
				}
				returnToBase(victim);
				return;
			}
			sendForRepair(victim);
			var gr=groupOfTank(victim);
			if (typeof(gr)=="undefined") { // defendGroup or aaGroup
				var list=enumGroup(defendGroup).concat(enumGroup(aaGroup));
				for (var i=0; i<list.length; ++i)
					orderDroidObj(list[i], DORDER_ATTACK, attacker);
				return;
			}
			// if coming for help, don't bother
			if (groupInfo[gr].helpPlayer!=-1 && distBetweenTwoPoints(attacker.x,attacker.y,groupInfo[gr].helpX,groupInfo[gr].helpY)>ARTILLERY_RANGE)
				return;
			groupInfo[gr].idleTime=0;
			if (!arty && !okToAllIn(groupInfo[gr].enemy))
				groupInfo[gr].delaystaticregroup = 1 + (attacker.type==DROID ? 1 : 0);
			// group comes for help when one of its droids is under attack
			var attackers=enumGroup(battleGroup[gr]);
			for (var i=0; i<attackers.length; ++i) 
				if (attackers[i].id!=victim.id)
					if (!sendForRepair(attackers[i]))
						if (attackers[i].order!=DORDER_SCOUT && attackers[i].order!=DORDER_ATTACK)
							orderDroidLoc(attackers[i], DORDER_SCOUT, attacker.x, attacker.y);
			return;
		} else { //Most likely a construction droid.
			//Demolish whatever the droid is building to save power (unless it's a base structure).
			if (victim.group==oilerGroup) {
				var list=enumStruct(me);
				var success=false;
				for (var i=0; i<list.length; ++i) {
					if ((list[i].status==BEING_BUILT) && distBetweenTwoPoints(victim.x,victim.y,list[i].x,list[i].y)<=HOTSPOT_SIZE && !isNearBase(list[i].x,list[i].y)) {
						orderDroidObj(victim, DORDER_DEMOLISH, list[i]);
						success=true;
						break;
					}
				}
				if (!success) // non-combat units retreat when attacked
					returnToBase(victim);
			}
		}
	} else if (victim.type == STRUCTURE) {
		if (arty) {
			lastArtyAttackX = victim.x;
			lastArtyAttackY = victim.y;
		}
		if (isBaseStructure(victim) && !arty)
			setStatus(STATUS_PANIC);
		if (attacker.player==scavengerPlayer && attacker.droidType==DROID_TRANSPORTER) { //A Scavenger Chopper is attacking one of our structures.
			if (victim.stattype == FACTORY)
				produceAATank(victim);
			if (victim.stattype == CYBORG_FACTORY)
				produceAACyborg(victim);
			if (status==STATUS_PANIC) //A Scavenger Chopper is attacking our base!
				attackedByChoppers=1;
		}
		// when a structure is attacked, all groups responsible for attacker's player react
		// unless the base itself is under attack, when all groups respond
		// (except on HARD or INSANE difficulties, since chances are that the main base would be doomed either way)
		for (var gr=0; gr<NUM_GROUPS; ++gr) 
			if (groupInfo[gr].enemy == attacker.player || status==STATUS_PANIC) {
				if (regroupWarriors(gr) && (status==STATUS_NORMAL || difficulty>=HARD)) 
					continue;
				// if coming for help, don't bother
				if (groupInfo[gr].helpPlayer!=-1 && distBetweenTwoPoints(attacker.x,attacker.y,groupInfo[gr].helpX,groupInfo[gr].helpY)>ARTILLERY_RANGE) 
					continue;
				var attackers=enumGroup(battleGroup[gr]);
				for (var i=0; i<attackers.length; ++i)
					if (!sendForRepair(attackers[i]))
						if (attackers[i].order!=DORDER_ATTACK)
							orderDroidObj(attackers[i], DORDER_ATTACK, attacker);
			}
		// defendGroup responds to all attacks as well
		var list=enumGroup(defendGroup);
		for (var i=0; i<list.length; ++i)
			orderDroidObj(list[i], DORDER_ATTACK, attacker);
		// aaGroup responds to attacks from enemy VTOLs
		if (attacker.type == DROID) {
			if (isVTOL(attacker) || attacker.droidType==DROID_TRANSPORTER || attacker.droidType==DROID_SUPERTRANSPORTER) {
				var list=enumGroup(aaGroup);
				for (var i=0; i<list.length; ++i)
					orderDroidObj(list[i], DORDER_ATTACK, attacker);
			}
		}
		// use VTOLs for defense as well
		for (var gr=0; gr<NUM_VTOL_GROUPS; ++gr) {
			var attackers=enumGroup(vtolGroup[gr]);
			for (var i=0; i<attackers.length; ++i)
				if (!sendForRepair(attackers[i]) && !isNaN(attackers[i].armed) && attackers[i].armed>=99)
					orderDroidObj(attackers[i], DORDER_ATTACK, attacker);
		}
		var attackers=enumGroup(abVtolGroup);
		for (var i=0; i<attackers.length; ++i)
			if (!sendForRepair(attackers[i]) && !isNaN(attackers[i].armed) && attackers[i].armed>=99)
				orderDroidObj(attackers[i], DORDER_ATTACK, attacker);
	}
}

//Something was just destroyed.
function eventDestroyed(object) {
	if (object.type == STRUCTURE) {
		if (object.stattype == HQ) {
			if (enumStruct(me,HQ).length==0) { //Our HQ was destroyed! Find another place to call home!
				findNewHomeBase();
				//Evacuate our old base and move to the new base.
				var droidList=enumDroid(me);
				for (var i=0; i<droidList.length; ++i) {
					if (isVTOL(droidList[i]))
						forceReturnToBase(droidList[i]); //Fly all our VTOLs to the new base.
					else if (typeof(droidList[i].group)!="undefined")
						if (droidList[i].group==builderGroup)
							forceReturnToBase(droidList[i]); //Send all our builderGroup trucks to the new base as well.
				}
			}
		}
	}
}

// we just picked up an oil drum (3.2+ only)
function eventPickup(feature, droid) {
	if (droid.player == me)
		returnToBase(droid);
}

// laser satellite strike
function eventStructureReady(structure) {
	list=[];
	for (var i=0; i<maxPlayers; ++i) if (isAnEnemy(i)) {
		list = list.concat(enumStruct(i),enumDroid(i));
	}
	ret=new naiveFindClusters(list,4); // 4 is the laser satellite splash radius
	// find an object closest to the middle of the biggest enemy cluster
	var minIdx=0,minDist=Infinity;
	for (var i=0; i<ret.maxCount; ++i) {
		var x1=ret.clusters[ret.maxIdx][i].x;
		var y1=ret.clusters[ret.maxIdx][i].y;
		var x2=ret.xav[ret.maxIdx];
		var y2=ret.yav[ret.maxIdx];
		var d=distBetweenTwoPoints(x1,y1,x2,y2);
		if (d<minDist) {
			minIdx=i;
			minDist=d;
		}
	}
	activateStructure(structure,list[ret.clusters[ret.maxIdx][minIdx]]);
}

// respond correctly on unit transfers
function eventObjectTransfer(object, from) {
	if (object.player != me)
		return; // object was transfered from me, not to me
	if (object.type == DROID) {
		eventDroidBuilt(object,null);
	} else {
		eventStructureBuilt(object,null);
	}
}

function eventGameInit() {
	setTimers();
}

// all initialization happens here, apart from personality initialization
function eventStartLevel() {
	personality=new constructPersonality();
	basePosition=startPositions[me];
	if (isStructureAvailable(fmod,me)) {
		HIGH_TECH_START = 1;
		EXTREME_LOW_POWER = 350;
	}
	for (var i=0; i<maxPlayers; ++i) {
		beaconInfo[i]=new constructBeaconInfo();
	}
	// HACK: find a better way to find out if we are in challenge
	if (playerPower(me)<1000 || playerPower(me) > 3300) 
		UNUSUAL_SITUATION = 1;
	becomeHarder();
	if (baseType != CAMP_CLEAN)
		personality.PEACE_TIME=0;
	var list = enumDroid(me);
	enemyInfo[-1]=new constructEnemyInfo();
	for (var i=0; i<maxPlayers; ++i) {
		enemyInfo[i]=new constructEnemyInfo();
		for (var j=0; j<list.length; ++j)
			if (!droidCanReach(list[j],startPositions[i].x,startPositions[i].y))
				enemyInfo[i].reachable=false;
	}
	builderGroup=newGroup();
	oilerGroup=newGroup();
	defendGroup=newGroup();
	aaGroup=newGroup();
	harassGroup=newGroup();
	aaVtolGroup=newGroup();
	abVtolGroup=newGroup();
	for (i=0; i<NUM_GROUPS; ++i)
		battleGroup[i]=newGroup();
	groupInfo[0]=new constructGroupInfo(findEnemy());
	for (i=1; i<NUM_GROUPS; ++i)
		groupInfo[i]=new constructGroupInfo(nextEnemy(groupInfo[i-1].enemy));
	NUM_VTOL_GROUPS = MIN_VTOL_GROUPS;
	for (i=0; i<MAX_VTOL_GROUPS; ++i)
		vtolGroup[i]=newGroup();
	for (var i=0; i<list.length; ++i) {
		var droid=list[i];
		if (droid.droidType == DROID_CONSTRUCT || droid.name.indexOf("Engineer")>-1) {
			addTruckToSomeGroup(droid);
			// the next line helps avoiding some strange problem when droids that are initially
			// buried into the ground don't move out of the way when a building is being
			// placed right above them
			orderDroidLoc(droid,DORDER_MOVE,droid.x+random(5)-2,droid.y+random(5)-2); 
		}
		else if (isVTOL(droid))
			addVtolToSomeGroup(droid);
		else if (droid.droidType == DROID_WEAPON || (droid.droidType == DROID_CYBORG && droid.name.indexOf("Mechanic")==-1))
			addTankToSomeGroup(droid);
	}
	hotspots[0]=new constructHotspotInfo(0,0);
	hotspots[0].lifespan=0;
	workWithAllies();
	setStatus(STATUS_EARLYGAME);
	adapt();
	setBuildOrder();
	queue("executeBuildOrder",200);
}

function eventChat(from, to, message) {
	// we are not case-sensitive
	message=message.toLowerCase();
	// introduce yourself
	if (message=="!nb who") {
		if (allianceExistsBetween(me,from)) {
			if (typeof(introduceYourselfToAlly)!="undefined")
				chat(ALLIES,introduceYourselfToAlly());
			else 
				if (typeof(introduceYourselfToEnemy)!="undefined")
					chat(ALLIES,introduceYourselfToEnemy());
			return;
		}
		if (typeof(introduceYourselfToEnemy)!="undefined")
			chat(from,introduceYourselfToEnemy());
		return;
	}
	// don't reply on any other message coming from enemies
	if (!allianceExistsBetween(from,me))
		return;
	// switch subpersonalities in run-time, if supported by the personality
	if (message.substr(0,7)=="!nb set") {
		if (typeof(switchSubpersonality)!="undefined") {
			chat(from,switchSubpersonality(message.substr(8)));
		}
		return;
	}
	if (message.substr(0,7)=="!nb res") {
		if (typeof(switchResearchPath)!="undefined") {
			chat(from,switchResearchPath(message.substr(8)));
		}
		return;
	}
	// understand when one of the allied NullBots is attacked by VTOLs, don't issue a warning twice
	if (message.indexOf("using vtols")>-1 && (message.indexOf("player")==0 || message.indexOf("scavenger")==0 || message.indexOf("enemy")==0)) {
		chat(ALLIES,"Got it.");
		if (MIN_DEFENDERS_AA<16)
			MIN_DEFENDERS_AA=16; //Start producing some anti-air units to protect our base.
		warnVtols=true;
		return;
	}
	// come for help when somebody is asking, even if it's Nexus AI
	if (message=="!nb help" || message=="help me!!" || message=="!nb go") {
		if (beaconInfo[from].exists == 0) {
			chat(from,"Please drop a beacon");
			return;
		}
		chat(from,sendHelp(from));
		return;
	}
	// stop sending units for help
	if (message=="!nb thanks" || message=="!nb tx") {
		chat(from,dontSendHelp(from));
		return;
	}
	// give a truck
	if (message=="!nb truck") {
		chat(from,giveTruck(from));
		return;
	}
	// give some money
	if (message=="!nb power" || message=="!nb money") {
		chat(from,giveMoney(from));
		return;
	}
}

function eventBeacon(x, y, from, to) {
	beaconInfo[from].x=x;
	beaconInfo[from].y=y;
	beaconInfo[from].exists=1;
	beaconInfo[from].lastUpdate=gameTime;
}

function eventBeaconRemoved(from, to) {
	beaconInfo[from].exists=0;
}

// currently unused events
function eventObjectSeen(viewer, seen) {}
